from searchbible import config
from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit.application import run_in_terminal
#from prompt_toolkit.application import get_app
import re


class RegexValidator(Validator):
    def __init__(self, pattern=""):
        super().__init__()
        self.filter = pattern # e.g. "^[0-9|]*?$"

    def validate(self, document):
        text = document.text

        if re.search(self.filter, text) or text.lower() in (config.exit_entry, config.cancel_entry):
            pass
        else:
            raise ValidationError(message='This entry accepts numbers only!', cursor_position=len(text)-1)


class NumberValidator(Validator):
    def validate(self, document):
        text = document.text

        if text.lower() in (config.exit_entry, config.cancel_entry):
            pass
        elif text and not text.isdigit():
            i = 0

            # Get index of first non numeric character.
            # We want to move the cursor here.
            for i, c in enumerate(text):
                if not c.isdigit():
                    break

            raise ValidationError(message='This entry accepts numbers only!', cursor_position=i)


class FloatValidator(Validator):
    def validate(self, document):
        text = document.text

        if text.lower() in (config.exit_entry, config.cancel_entry):
            pass
        try:
            float(text)
        except:
            raise ValidationError(message='This entry accepts floating point numbers only!', cursor_position=0)


class NoAlphaValidator(Validator):
    def validate(self, document):
        text = document.text

        if text.lower() in (config.exit_entry, config.cancel_entry):
            pass
        elif text and text.isalpha():
            i = 0

            # Get index of first non numeric character.
            # We want to move the cursor here.
            for i, c in enumerate(text):
                if c.isalpha():
                    break

            raise ValidationError(message='This entry does not accept alphabet letters!', cursor_position=i)


"""
Modified from LetMeDoIt AI Plugin - bible studies

add input suggestions and predefined instructions for bible studies

"""

bible_study_suggestions = []

bible_study_suggestions.append("""Bible Topic.
Write about the following topic in reference to the Bible.
In addition, explain the significance of the topic in the bible.
I want your writing to be comprehensive and informative.
Remember, in your writing, please provide me with concrete examples from the Bible and the bible references from the text you are citing.

# Bible topic
""")

bible_study_suggestions.append("""Bible Place.
I will provide you with a location name.
Give me comprehensive information on this location in the bible.
If this singular name is used to denote various different locations in the Bible, kindly provide separate descriptions for each one.
Explain the significance of this location in the bible.

# Bible location
""")

bible_study_suggestions.append("""Bible Person.
I will provide you with a person name.
Give me comprehensive information on this person in the bible.
If this singular name is used to denote various different characters in the Bible, kindly provide separate descriptions for each one.
Explain the significance of this person in the bible.

# Bible preson name
""")

# Translation
bible_study_suggestions.append("""Bible Hebrew Verse Translation.
I would like to request your assistance as a bible translator.
I will provide you with a Hebrew bible verse text.
First, give me transliteration of the text.
Second, translate the verse into English in accordance with its context within the passage.
Remember, translate from the Hebrew text that I give you, do not copy an existing translation, as I want a completely new translation.
Third, map each Hebrew word with corresponding translation.
In mapping section, do include transliteration of each Hebrew word, so that each mapping are formatted in pattern this pattern: word | transliteration | corresponding translation.
Use "Transliteration:", "Translation:" and "Mapping:" as the section titles.
Remember, do not give parsing information, explanation or repeat the bible reference that I give you.

I am giving you the Hebrew verse text below:

# Hebrew verse
""")

bible_study_suggestions.append("""Bible Greek Verse Translation.
I would like to request your assistance as a bible translator.
I will provide you with a Greek bible verse text.
First, give me transliteration of the text.
Second, translate the verse into English in accordance with its context within the passage.
Remember, translate from the Greek text that I give you, do not copy an existing translation, as I want a completely new translation.
Third, map each Greek word with corresponding translation.
In mapping section, do include transliteration of each Greek word, so that each mapping are formatted in pattern this pattern: word | transliteration | corresponding translation.
Use "Transliteration:", "Translation:" and "Mapping:" as the section titles.
Remember, do not give parsing information, explanation or repeat the bible reference that I give you.

I am giving you the Greek verse text below:

# Greek verse
""")

# Book Features
bible_study_suggestions.append("""Bible Book Introduction.
Write a detailed introduction on a book in the bible, considering all the following questions:
1. Who is the author or attributed author of the book?
2. What is the date or time period when the book was written?
3. What is the main theme or purpose of the book?
4. What are the significant historical events or context surrounding the book?
5. Are there any key characters or figures in the book?
6. What are some well-known or significant passages from the book?
7. How does the book fit into the overall structure and narrative of the Bible?
8. What lessons or messages can be learned from the book?
9. What is the literary form or genre of the book (e.g. historical, prophetic, poetic, epistle, etc.)?
10. Are there any unique features or controversies surrounding the book?
I want the introduction to be comprehensive and informative.
When you explain, quote specific words or phases from relevant bible verses, if any.
Answer all these relevant questions mentioned above, in the introduction, pertaining to the following bible book.

# Bible book name
""")

bible_study_suggestions.append("""Bible Outline.
I am currently studying the following bible book and chapters and would appreciate it if you could provide me with a detailed outline for it.
I want an outline for all chapters of the bible book and chapters I am giving you.
Please divide it into several main sections and under each section, divide it into different passages.
Break down the passages into the smallest ones possible, as I am looking for a highly detailed outline.
Additionally, kindly provide a title for each passage.
Remember, use your own analysis based on the bible text instead of copying a published outline.
If chapters are not specified, provide an outline to cover all chapters of the bible book given.

# Bible reference
""")

# Chapter Features
bible_study_suggestions.append("""Bible Chapter Summary.
Write a detailed interpretation on a bible chapter, considering all the following questions:
1. What is the overview of the chapter?
2. How are the verses in this chapter structured or organized?
3. Are there any key verses or passages in the chapter?
4. Are there any significant characters, events, or symbols in the chapter?
5. What is the main themes or messages of the chapter?
6. What historical or cultural context is important to understand the chapter?
7. How have theologians, scholars, or religious leaders interpreted this chapter?
8. Are there any popular interpretations or controversies related to this chapter?
9. How does this chapter relate to other chapters, books, or themes in the Bible?
10. What lessons or morals can be taken from the chapter?
I want your interpretation to be comprehensive and informative.
When you explain, quote specific words or phases from relevant bible verses, if any.
Answer all these relevant questions mentioned above, in the interpretation, pertaining to the following bible chapter.

# Bible chapter
""")

# Verse Features
bible_study_suggestions.append("""Bible OT Verse Interpretation.
Interpret the following verse in the light of its context, together with insights of biblical Hebrew studies.
I want your interpretation to be comprehensive and informative.  When you explain, quote specific words or phases from the verse.
However, remember, I want you not to quote the whole verse word by word, especially in the beginning of your response, as I already know its content.

# Bible verse
""")

bible_study_suggestions.append("""Bible NT Verse Interpretation.
Interpret the following verse in the light of its context, together with insights of biblical Greek studies.
I want your interpretation to be comprehensive and informative.  When you explain, quote specific words or phases from the verse.
However, remember, I want you not to quote the whole verse word by word, especially in the beginning of your response, as I already know its content.

# Bible verse
""")

bible_study_suggestions.append("""Bible Insights.
Give me exegetical insights in detail on the following bible verses:

# Verses
""")

# Studies
bible_study_suggestions.append("""Bible Perspective.
Tell me how we should understand the content given below, according to biblical perspectives and principles:

# Content
""")

bible_study_suggestions.append("""Bible Meaning.
Explain the meaning of the content given below in reference to the Bible:

# Content
""")

bible_study_suggestions.append("""Bible Summary.
Write a summary on the content given below in reference to the Bible:

# Content
""")

bible_study_suggestions.append("""Bible Highlights in an Old Testament passage in the bible.
Can you give me a detailed summary of this passage?
How can the meaning of the biblical Hebrew words in this passage provide insights that aid our understanding? Explain the insights with examples.
What are the key words and phrases used in this passage, and how do they contribute to its meaning?
How does this passage relate to other passages in the Bible, both within the same book and in other books?
Please answer all relevant questions pertaining to the following passage.
You don't need to provide all the verses from the passage, as I am already familiar with it. 
I want your response focus on my questions.  However, do not repeat my questions word by word in your answer.

Please write in detail pertaining to the following passage:

# Passage
""")

bible_study_suggestions.append("""Bible Highlights in a New Testament passage in the bible.
Can you give me a detailed summary of this passage?
How can the meaning of the biblical Greek words in this passage provide insights that aid our understanding? Explain the insights with examples.
What are the key words and phrases used in this passage, and how do they contribute to its meaning?
How does this passage relate to other passages in the Bible, both within the same book and in other books?
Please answer all relevant questions pertaining to the following passage.
You don't need to provide all the verses from the passage, as I am already familiar with it. 
I want your response focus on my questions.  However, do not repeat my questions word by word in your answer.

Please write in detail pertaining to the following passage:

# Passage
""")

bible_study_suggestions.append("""Bible Historical Context of an Old Testament passage in the bible.
In what ways does this passage speak to the struggles that Old Testament People faced in the time of its writing, and how does an understanding of its historical and cultural context help us to better interpret its meaning?
In what ways does a deeper comprehension of the culture of the Ancient Near East aid in our understanding of the passage? Give examples based on the content of the given passage.
How does comprehending the social, political, and religious environment of the Old Testament era, along with the distinct obstacles that the people of that time may have encountered in their daily lives, contribute to our interpretation of the given passage? Give examples based on the content of the given passage.
Please answer all relevant questions pertaining to the following passage.
Do not give me general message of the passage, as I am seeking specific insights as I am seeking specific insights about Old Testament People stugglings in view of historical context.
I already know the content of the passage.  Please do not repeat.  
Do not give me summary of the passage that is not relevant to the struggling or historical context.

Please write in detail pertaining to the following passage:

# Passage
""")

bible_study_suggestions.append("""Bible Historical Context of a New Testament passage in the bible.
In what ways does this passage speak to the struggles that early Christians faced in the time of its writing, and how does an understanding of its historical and cultural context help us to better interpret its meaning?
In what ways does a deeper comprehension of the Jewish culture aid in our understanding of the passage? Give examples based on the content of the given passage.
How does comprehending the social, political, and religious environment of the New Testament era, along with the distinct obstacles that the people of that time may have encountered in their daily lives, contribute to our interpretation of the given passage? Give examples based on the content of the given passage.
Please answer all relevant questions pertaining to the following passage.
Do not give me general message of the passage, as I am seeking specific insights as I am seeking specific insights about Christian stugglings in view of historical context.
I already know the content of the passage.  Please do not repeat.  
Do not give me summary of the passage that is not relevant to the struggling or historical context.

Please write in detail pertaining to the following passage:

# Passage
""")

bible_study_suggestions.append("""Bible Themes in an Old Testament passage.
What are the key themes expressed in this passage?  Please elaborate on each theme in details.  
Explain how biblical Hebrew studies help us understand each theme.  Povide illustration that aid our understanding, if possible.
What are the theological implications of this passage, and how does it contribute to our understanding of God's character and plan for humanity?
How does the New Testament contribute to our comprehension of this Old Testament passage? Give examples and quote related New Testament passages.
How does the message conveyed in this passage impact our connection with God?
Please answer all relevant questions pertaining to the following passage.
Do not give me historical context of the passage, as I already know them. 
Do not give me general information about the passage, as I am seeking specific themes, theological implications and connection with God.

Please answer all relevant questions pertaining to the following passage:

# Passage
""")

bible_study_suggestions.append("""Bible Themes in a New Testament passage.
What are the key themes expressed in this passage?  Please elaborate on each theme in details.  
Explain how biblical Greek studies help us understand each theme.  Povide illustration that aid our understanding, if possible.
What are the theological implications of this passage, and how does it contribute to our understanding of God's character and plan for humanity?
How does the Old Testament contribute to our comprehension of this New Testament passage? Give examples and quote related Old Testament passages.
How does the message conveyed in this passage impact our connection with God?
Please answer all relevant questions pertaining to the following passage.
Do not give me historical context of the passage, as I already know them. 
Do not give me general information about the passage, as I am seeking specific themes, theological implications and connection with God.

Please answer all relevant questions pertaining to the following passage:

# Passage
""")

bible_study_suggestions.append("""Bible Application.
Provide detailed applications of a bible passages, without general introduction or summary.

Please address the following questions in your answer:
How can I relate this passage to my own daily life and personal circumstances?
How can I apply the principles and teachings of this passage to my relationships with God and others?
What can I do to apply the lessons from this passage in my daily life?
Are there any practical steps or behaviors that I can develop or change based on this passage?

Please answer all relevant questions pertaining to the following passage:

# Passage
""")

bible_study_suggestions.append("""Bible Sermon.
Write a sermon with the following questions in mind:
1. What is the context of the passage? Who wrote it, to whom was it written, and what historical or cultural background is important to know?
2. What is the main message or theme of the passage? How does it relate to the overall message of the Bible?
3. What is the message or lesson that you want the congregation to take away from the sermon?
4. How can I make the sermon relevant and applicable to the lives of the congregation members?
5. What are some relevant stories, personal experiences, or examples that can be used to illustrate the main points of the sermon?
6. What are some potential challenges or objections that someone might have to the message of the sermon, and how can those be addressed?
7. Are there any important theological or doctrinal principles that need to be emphasized or clarified in relation to the passage?

Please write pertaining to the following passage:

# Passage
""")

bible_study_suggestions.append("""Bible Devotion.
Write a devotion on a bible passage, based on the following guidance:
1. Get a full understanding of its meaning and context.
2. Reflect on how the passage applies to your own life and experiences.
3. Consider what message or lesson God may be trying to communicate through the passage.
4. Decide on a specific theme or topic you want to focus on in your devotion, based on the Bible passage.
5. Use personal anecdotes, relevant scripture references, and practical applications to help illustrate and reinforce your message.
6. End your devotion with a call to action or a prayer that encourages readers to apply the lesson to their own lives.
7. Edit and revise your devotion to ensure that it is clear, concise, and impactful.
I am already familiar with the contents of the passage, please refrain from providing general introduction or summary.
I want deep insights about devotions.

Please write pertaining to the following passage:

# Passage
""")

bible_study_suggestions.append("""Bible Prayer.
Write a prayer pertaining to the following content in reference to the Bible:

# Content
""")

bible_study_suggestions.append("""Bible Short Prayer.
Write a prayer pertaining to the following content in reference to the Bible.
(Keep the prayer short and do not make it longer than a paragraph.)

# Content
""")

bible_study_suggestions.append("""Bible Review Questions.
Assist me to prepare materials for leading a bible study group.
I'm looking for specific questions that focus on a particular passage, rather than general bible study questions. Could you help me with that?
I'm hoping to facilitate a meaningful discussion and deeper understanding of this passage through targeted questions. Could you guide me in creating those questions?
Remember, I am already familiar with the contents of the passage, please refrain from providing me with general information or summary. Please give me review questions for bible studies directly.

Please answer all relevant questions pertaining to the following passage:

# Passage
""")

bible_study_suggestions.append("""Bible Key Words.
Identify key words in the content given below.
Elaborate on their importance in comprehending the context and the bible as a whole.
I want your elaboration to be comprehensive and informative.
Remember, in your writing, please provide me with concrete examples from the Bible and the bible references from the text you are citing.

# Content
""")

bible_study_suggestions.append("""Bible Key Themes.
Identify key themes or key messages in the content given below.
Elaborate on their importance in comprehending the content given below and the bible as a whole.
I want your elaboration to be comprehensive and informative.
Remember, in your writing, please provide me with concrete examples from the Bible and the bible references from the text you are citing.

# Content
""")

bible_study_suggestions.append("""Bible Theology.
Please write the theological messages conveyed in the content given below, in reference to the Bible.
In addition, explain the significance of the theological messages in the bible.
I want your writing to be comprehensive and informative.
Remember, in your writing, please provide me with concrete examples from the Bible and the bible references from the text you are citing.

# Content
""")

bible_study_suggestions.append("""Bible Promises.
Quote relevant Bible promises in relation to the content given below.
Explain how these promises are meant to provide comfort, hope, and guidance in my life.

# Content
""")

bible_study_suggestions.append("""Bible Passages.
Quote relevant Bible passages in relation to the content given below given below. Explain how these passages are related to the given content.

# Content
""")

bible_study_suggestions.append("""Bible Thought Progression.
Bible Thought Progression of a bible book / chapter / passage.
Provide a detailed overview of the author's thought progression in effectively conveying the key messages of the following book / chapter / passage:

# Book / Chapter / Passage
""")

bible_study_suggestions.append("""Bible Canonical Context.
Bible Canonical Context of a bible book / chapter / passage.
How is this book / chapter / passage connected or linked with other books in the Bible?
What is the canonical context of this given book / chapter / passage in light of the whole canon of the bible?
How does this book / chapter / passage fit into the larger Biblical story in light of the whole canon of the bible?
                               
Please answer all relevant questions pertaining to the following book / chapter / passage with concrete examples:

# Book / Chapter / Passage
""")



import pydoc, textwrap
from searchbible import config
from searchbible.utils.prompt_shared_key_bindings import prompt_shared_key_bindings
from searchbible.utils.prompt_multiline_shared_key_bindings import prompt_multiline_shared_key_bindings
from searchbible.utils.prompt_validator import NumberValidator
from prompt_toolkit import prompt
from prompt_toolkit.application import run_in_terminal
from prompt_toolkit.styles import Style
from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding import KeyBindings, merge_key_bindings, ConditionalKeyBindings
from prompt_toolkit.shortcuts import set_title
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard

class Prompts:

    def __init__(self, custom_key_bindings=None):
        self.custom_key_bindings = custom_key_bindings
        config.multilineInput = False
        config.clipboard = PyperclipClipboard()
        config.promptStyle1 = self.promptStyle1 = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor1,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor1,
        })
        config.promptStyle2 = self.promptStyle2 = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor2,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor2,
        })
        self.inputIndicator = [
            ("class:indicator", ">>> "),
        ]
        self.shareKeyBindings()
        config.showKeyBindings = self.showKeyBindings
        config.terminalColors = {
            "ansidefault": "ansidefault",
            "ansiblack": "ansiwhite",
            "ansired": "ansibrightred",
            "ansigreen": "ansibrightgreen",
            "ansiyellow": "ansibrightyellow",
            "ansiblue": "ansibrightblue",
            "ansimagenta": "ansibrightmagenta",
            "ansicyan": "ansibrightcyan",
            "ansigray": "ansibrightblack",
            "ansiwhite": "ansiblack",
            "ansibrightred": "ansired",
            "ansibrightgreen": "ansigreen",
            "ansibrightyellow": "ansiyellow",
            "ansibrightblue": "ansiblue",
            "ansibrightmagenta": "ansimagenta",
            "ansibrightcyan": "ansicyan",
            "ansibrightblack": "ansigray",
        }

    def getToolBar(self, multiline=False):
        if multiline:
            return f" [ctrl+q] exit [escape+enter] send "
        return f" [ctrl+q] exit [enter] send "

    def shareKeyBindings(self):
        this_key_bindings = KeyBindings()
        @this_key_bindings.add("c-q")
        def _(event):
            buffer = event.app.current_buffer
            buffer.text = config.exit_entry
            buffer.validate_and_handle()
        @this_key_bindings.add("escape", "d")
        def _(_):
            config.developer = not config.developer
            config.saveConfig()
            run_in_terminal(lambda: config.print3(f"Developer mode: {'enabled' if config.developer else 'disabled'}!"))
        @this_key_bindings.add("c-l")
        def _(_):
            config.toggleMultiline()
        @this_key_bindings.add("c-k")
        def _(_):
            run_in_terminal(self.showKeyBindings)
        @this_key_bindings.add("escape", "r")
        def _(_):
            print(f"Restarting Search Bible AI ...")
            config.restartApp()
        @this_key_bindings.add("c-w")
        def _(_):
            config.wrapWords = not config.wrapWords
            config.saveConfig()
            run_in_terminal(lambda: config.print3(f"Word Wrap: '{'enabled' if config.wrapWords else 'disabled'}'!"))
        @this_key_bindings.add("escape", "m")
        def _(_):
            config.mouseSupport = not config.mouseSupport
            config.saveConfig()
            run_in_terminal(lambda: config.print3(f"Entry Mouse Support: '{'enabled' if config.mouseSupport else 'disabled'}'!"))

        conditional_prompt_multiline_shared_key_bindings = ConditionalKeyBindings(
            key_bindings=prompt_multiline_shared_key_bindings,
            filter=Condition(lambda: config.multilineInput),
        )
        available_key_bindings = [
            this_key_bindings,
            conditional_prompt_multiline_shared_key_bindings,
            prompt_shared_key_bindings,
        ]
        if self.custom_key_bindings:
            available_key_bindings.append(self.custom_key_bindings)
        self.prompt_shared_key_bindings = merge_key_bindings(available_key_bindings)

    def showKeyBindings(self):
        bindings = {
            "ctrl+q": "quit / exit current feature",
            "ctrl+s": "toggle chapter subheadings",
            "ctrl+y": "play bible audio",
            "ctrl+f": "search verses",
            "esc+f": "search paragraphs",
            "ctrl+p": "toggle bible comparison",
            "esc+p": "select bibles for comparison",
            "ctrl+g": "launch ChatGPT prompt",
            "esc+g": "launch GeminiPro prompt",
            "esc+c": "change configs",
            "ctrl+z": "cancel typing",
            "ctrl+a": "select / unselect all",
            "ctrl+c": "copy [w/ mouse support]",
            "ctrl+v": "paste [w/ mouse support]",
            "ctrl+x": "cut [w/ mouse support]",
            #"ctrl+n": "new chat",
            #"ctrl+y": "new chat without context",
            #"ctrl+s": "save chat",
            #"ctrl+o": "change predefined context",
            #"ctrl+g": "pager view",
            "ctrl+d": "forward delete",
            "ctrl+h": "backspace",
            #"ctrl+e": "swap command execution mode",
            "ctrl+k": "show key bindings",
            "ctrl+l": "toggle multi-line entry",
            #"ctrl+b": "toggle input audio",
            #"ctrl+p": "toggle response audio",
            "ctrl+w": "toggle word wrap",
            "ctrl+r": "insert a linebreak",
            #"ctrl+i or tab": "insert a file or folder path",
            #"shift+tab": f"insert '{config.terminalEditorTabText}' [configurable]",
            #"esc+f": "display device information",
            #"esc+c": "count current message tokens",
            #"esc+i": "toggle improved writing feature",
            "esc+m": "toggle mouse support",
            #"esc+t": "system command prompt",
            "esc+b": "move cursor to line beginning",
            "esc+e": "move cursor to line end",
            "esc+a": "move cursor to entry beginning",
            "esc+z": "move cursor to entry end",
            "esc+s": "swap text brightness",
            "esc+d": "swap developer mode",
            "esc+r": "restart SearchBibleAI",
        }
        #textEditor = config.customTextEditor.split(" ", 1)[0]
        #bindings["ctrl+e"] = f"""edit current input with '{config.customTextEditor if textEditor and config.isPackageInstalled(textEditor) else "eTextEdit"}'"""
        #bindings["esc+p"] = f"""edit the previous response with '{config.customTextEditor if textEditor and config.isPackageInstalled(textEditor) else "eTextEdit"}'"""
        multilineBindings = {
            "enter": "new line",
            "esc+enter": "complete entry",
            "esc+1": "go up 10 lines",
            "esc+2": "go up 20 lines",
            "esc+3": "go up 30 lines",
            "esc+4": "go up 40 lines",
            "esc+5": "go up 50 lines",
            "esc+6": "go up 60 lines",
            "esc+7": "go up 70 lines",
            "esc+8": "go up 80 lines",
            "esc+9": "go up 90 lines",
            "esc+0": "go up 100 lines",
            "f1": "go down 10 lines",
            "f2": "go down 20 lines",
            "f3": "go down 30 lines",
            "f4": "go down 40 lines",
            "f5": "go down 50 lines",
            "f6": "go down 60 lines",
            "f7": "go down 70 lines",
            "f8": "go down 80 lines",
            "f9": "go down 90 lines",
            "f10": "go down 100 lines",
            "ctrl+u": f"go up '{config.terminalEditorScrollLineCount}' lines [configurable]",
            "ctrl+j": f"go down '{config.terminalEditorScrollLineCount}' lines [configurable]",
        }
        keyHelp = f"{config.divider}\n"
        keyHelp += "# Key Bindings\n"
        keyHelp += "[blank]: launch action menu\n"
        for key, value in bindings.items():
            keyHelp += f"{key}: {value}\n"
        keyHelp += "\n## Key Bindings\n[for multiline entry only]\n"
        for key, value in multilineBindings.items():
            keyHelp += f"{key}: {value}\n"
        keyHelp += f"{config.divider}\n"

        if config.isPackageInstalled("less"):
            pydoc.pipepager(f"To close this help page, press 'q'\n\n{keyHelp}\nTo close this help page, press 'q'", cmd='less -R')
        else:
            print(keyHelp)

    def simplePrompt(self, numberOnly=False, validator=None, inputIndicator="", default="", accept_default=False, completer=None, promptSession=None, style=None, is_password=False, bottom_toolbar=None):
        config.selectAll = False
        inputPrompt = promptSession.prompt if promptSession is not None else prompt
        if not inputIndicator:
            inputIndicator = self.inputIndicator
        if numberOnly:
            validator = NumberValidator()
        userInput = inputPrompt(
            inputIndicator,
            key_bindings=self.prompt_shared_key_bindings,
            bottom_toolbar=self.getToolBar(config.multilineInput) if bottom_toolbar is None else bottom_toolbar,
            #enable_system_prompt=True,
            swap_light_and_dark_colors=Condition(lambda: not config.terminalResourceLinkColor.startswith("ansibright")),
            style=self.promptStyle1 if style is None else style,
            validator=validator,
            multiline=Condition(lambda: config.multilineInput),
            default=default,
            accept_default=accept_default,
            completer=completer,
            is_password=is_password,
            mouse_support=Condition(lambda: config.mouseSupport),
            clipboard=config.clipboard,
        )
        userInput = textwrap.dedent(userInput) # dedent to work with code block
        return userInput if hasattr(config, "addPathAt") and config.addPathAt else userInput.strip()



import os, traceback, json, pprint, wcwidth, textwrap, threading, time, subprocess, sys, re, pkg_resources, requests, shutil
import openai, tiktoken
from openai import OpenAI
from pygments.styles import get_style_by_name
from prompt_toolkit.styles.pygments import style_from_pygments_cls
from prompt_toolkit import print_formatted_text, HTML
from prompt_toolkit import prompt
from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding import KeyBindings, merge_key_bindings
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
from searchbible.utils.prompt_shared_key_bindings import prompt_shared_key_bindings
from searchbible import config
from searchbible.utils.shortcuts import createShortcuts
from pathlib import Path
from packaging import version


def check_openai_errors(func):
    def wrapper(*args, **kwargs):
        try: 
            return func(*args, **kwargs)
        except openai.APIError as e:
            print("Error: Issue on OpenAI side.")
            print("Solution: Retry your request after a brief wait and contact us if the issue persists.")
        except openai.APIConnectionError as e:
            print("Error: Issue connecting to our services.")
            print("Solution: Check your network settings, proxy configuration, SSL certificates, or firewall rules.")
        except openai.APITimeoutError as e:
            print("Error: Request timed out.")
            print("Solution: Retry your request after a brief wait and contact us if the issue persists.")
        except openai.AuthenticationError as e:
            print("Error: Your API key or token was invalid, expired, or revoked.")
            print("Solution: Check your API key or token and make sure it is correct and active. You may need to generate a new one from your account dashboard.")
            HealthCheck.changeAPIkey()
        except openai.BadRequestError as e:
            print("Error: Your request was malformed or missing some required parameters, such as a token or an input.")
            print("Solution: The error message should advise you on the specific error made. Check the [documentation](https://platform.openai.com/docs/api-reference/) for the specific API method you are calling and make sure you are sending valid and complete parameters. You may also need to check the encoding, format, or size of your request data.")
        except openai.ConflictError as e:
            print("Error: The resource was updated by another request.")
            print("Solution: Try to update the resource again and ensure no other requests are trying to update it.")
        except openai.InternalServerError as e:
            print("Error: Issue on OpenAI servers.")
            print("Solution: Retry your request after a brief wait and contact us if the issue persists. Check the [status page](https://status.openai.com).")
        except openai.NotFoundError as e:
            print("Error: Requested resource does not exist.")
            print("Solution: Ensure you are the correct resource identifier.")
        except openai.PermissionDeniedError as e:
            print("Error: You don't have access to the requested resource.")
            print("Solution: Ensure you are using the correct API key, organization ID, and resource ID.")
        except openai.RateLimitError as e:
            print("Error: You have hit your assigned rate limit.")
            print("Solution: Pace your requests. Read more in OpenAI [Rate limit guide](https://platform.openai.com/docs/guides/rate-limits).")
        except openai.UnprocessableEntityError as e:
            print("Error: Unable to process the request despite the format being correct.")
            print("Solution: Please try the request again.")
        except:
            print(traceback.format_exc())
    return wrapper

class HealthCheck:

    # token limit
    # reference: https://platform.openai.com/docs/models/gpt-4
    tokenLimits = {
        #"gpt-3.5-turbo-instruct": 4097,
        "gpt-3.5-turbo": 4097,
        "gpt-3.5-turbo-16k": 16385,
        "gpt-3.5-turbo-1106": 16385,
        "gpt-4-turbo-preview": 128000,
        "gpt-4-0125-preview": 128000,
        "gpt-4-1106-preview": 128000, # official 128,000; but "This model supports at most 4096 completion tokens"; set 8192 here to work with LetMeDoIt AI dynamic token feature
        #"gpt-4-vision-preview": 128,000, # used in plugin "analyze images"
        "gpt-4": 8192,
        "gpt-4-32k": 32768,
    }

    models = tuple(tokenLimits.keys())

    @staticmethod
    def check():
        HealthCheck.checkPythonVersion()
        if config.autoUpgrade:
            HealthCheck.updateApp()
        if not hasattr(config, "exit_entry"):
            HealthCheck.setBasicConfig()
        createShortcuts()
        HealthCheck.setOsOpenCmd()
        #HealthCheck.checkPygame()

    @staticmethod
    def setBasicConfig(): # minimum config to work with standalone scripts built with AutoGen
        config.openaiApiKey = ''
        config.chatGPTApiModel = 'gpt-3.5-turbo-16k'
        config.chatGPTApiMaxTokens = 4000
        config.chatGPTApiMinTokens = 256
        config.llmTemperature = 0.8
        config.max_consecutive_auto_reply = 10
        config.exit_entry = '.exit'
        config.cancel_entry = '.cancel'
        config.terminalPromptIndicatorColor1 = "ansimagenta"
        config.terminalPromptIndicatorColor2 = "ansicyan"
        config.terminalCommandEntryColor1 = "ansiyellow"
        config.terminalCommandEntryColor2 = "ansigreen"
        config.terminalResourceLinkColor = "ansiyellow"
        config.terminalHeadingTextColor = "ansigreen"
        config.mouseSupport = False
        config.embeddingModel = "paraphrase-multilingual-mpnet-base-v2"
        config.max_agents = 5
        config.max_group_chat_round = 12
        config.max_consecutive_auto_reply = 10
        config.includeIpInSystemMessage = False
        config.wrapWords = True
        config.pygments_style = ""
        # work with bible verse parser
        config.noOfLinesPerChunkForParsing = 100
        config.parseEnglishBooksOnly = False
        config.useLiteVerseParsing = False
        config.standardAbbreviation = "ENG"
        config.convertChapterVerseDotSeparator = True
        config.parseBookChapterWithoutSpace = True
        config.parseBooklessReferences = True
        config.parseClearSpecialCharacters = False
        config.parserStandarisation = "NO"

    @staticmethod
    def setOsOpenCmd():
        if config.terminalEnableTermuxAPI:
            config.open = "termux-share"
        elif config.thisPlatform == "Linux":
            config.open = "xdg-open"
        elif config.thisPlatform == "Darwin":
            config.open = "open"
        elif config.thisPlatform == "Windows":
            config.open = "start"
        # name macOS
        if config.thisPlatform == "Darwin":
            config.thisPlatform = "macOS"

    @staticmethod
    def checkPygame():
        # try import pygame
        try:
            # hide pygame welcome message
            os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "1"
            import pygame
            pygame.mixer.init()
            config.isPygameInstalled = True
        except:
            config.isPygameInstalled = False

    # automatic update
    config.pipIsUpdated = False
    def updateApp():
        thisPackage = "searchbible"
        #thisPackage = f"{package}_android" if config.isTermux else thisPackage
        print(f"Checking '{thisPackage}' version ...")
        installed_version = HealthCheck.getPackageInstalledVersion(thisPackage)
        if installed_version is None:
            print("Installed version information is not accessible!")
        else:
            print(f"Installed version: {installed_version}")
        latest_version = HealthCheck.getPackageLatestVersion(thisPackage)
        if latest_version is None:
            print("Latest version information is not accessible at the moment!")
        elif installed_version is not None:
            print(f"Latest version: {latest_version}")
            if latest_version > installed_version:
                if config.thisPlatform == "Windows":
                    print("Automatic upgrade feature is yet to be supported on Windows!")
                    print(f"Run 'pip install --upgrade {thisPackage}' to manually upgrade this app!")
                else:
                    try:
                        # delete old shortcut files
                        appName = "SearchBibleAI"
                        shortcutFiles = (f"{appName}.bat", f"{appName}.command", f"{appName}.desktop")
                        for shortcutFile in shortcutFiles:
                            shortcut = os.path.join(config.packageFolder, shortcutFile)
                            if os.path.isfile(shortcut):
                                os.remove(shortcut)
                        # upgrade package
                        HealthCheck.installmodule(f"--upgrade {thisPackage}")
                        HealthCheck.restartApp()
                    except:
                        print(f"Failed to upgrade '{thisPackage}'!")

    @staticmethod
    def getPackageInstalledVersion(package):
        try:
            installed_version = pkg_resources.get_distribution(package).version
            return version.parse(installed_version)
        except pkg_resources.DistributionNotFound:
            return None

    @staticmethod
    def getPackageLatestVersion(package):
        try:
            response = requests.get(f"https://pypi.org/pypi/{package}/json", timeout=10)
            latest_version = response.json()['info']['version']
            return version.parse(latest_version)
        except:
            return None

    @staticmethod
    def installmodule(module, update=True):
        #executablePath = os.path.dirname(sys.executable)
        #pippath = os.path.join(executablePath, "pip")
        #pip = pippath if os.path.isfile(pippath) else "pip"
        #pip3path = os.path.join(executablePath, "pip3")
        #pip3 = pip3path if os.path.isfile(pip3path) else "pip3"

        isInstalled, _ = subprocess.Popen("pip -V", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
        pipInstallCommand = f"{sys.executable} -m pip install"

        if isInstalled:

            if update:
                if not config.pipIsUpdated:
                    pipFailedUpdated = "pip tool failed to be updated!"
                    try:
                        # Update pip tool in case it is too old
                        updatePip = subprocess.Popen(f"{pipInstallCommand} --upgrade pip", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                        *_, stderr = updatePip.communicate()
                        if not stderr:
                            print("pip tool updated!")
                        else:
                            print(pipFailedUpdated)
                    except:
                        print(pipFailedUpdated)
                    config.pipIsUpdated = True
            try:
                upgrade = (module.startswith("-U ") or module.startswith("--upgrade "))
                if upgrade:
                    moduleName = re.sub("^[^ ]+? (.+?)$", r"\1", module)
                else:
                    moduleName = module
                print(f"{'Upgrading' if upgrade else 'Installing'} '{moduleName}' ...")
                installNewModule = subprocess.Popen(f"{pipInstallCommand} {module}", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                *_, stderr = installNewModule.communicate()
                if not stderr:
                    print(f"Package '{moduleName}' {'upgraded' if upgrade else 'installed'}!")
                else:
                    print(f"Failed {'upgrading' if upgrade else 'installing'} package '{moduleName}'!")
                    if config.developer:
                        print(stderr)
                return True
            except:
                return False

        else:

            print("pip command is not found!")
            return False

    @staticmethod
    def checkPythonVersion():
        # requires python 3.8+; required by package 'tiktoken'
        pythonVersion = sys.version_info
        if pythonVersion < (3, 8):
            print("Python version higher than 3.8 is required!")
            print("Closing ...")
            exit(1)
        elif pythonVersion >= (3, 12):
            print("Some features may not work with python version newer than 3.11!")

    @staticmethod
    def setSharedItems():
        config.storagedirectory = HealthCheck.getFiles()
        config.excludeConfigList = []
        config.divider = "--------------------"
        # share frequently used methods in config to avoid circular imports
        if not hasattr(config, "print"):
            config.print = HealthCheck.print
        if not hasattr(config, "print2"):
            config.print2 = HealthCheck.print2
        if not hasattr(config, "print3"):
            config.print3 = HealthCheck.print3
        if not hasattr(config, "print4"):
            config.print4 = HealthCheck.print4
        if not hasattr(config, "restartApp"):
            config.restartApp = HealthCheck.restartApp
        if not hasattr(config, "isPackageInstalled"):
            config.isPackageInstalled = HealthCheck.isPackageInstalled
        if not hasattr(config, "saveConfig"):
            config.saveConfig = HealthCheck.saveConfig

    @staticmethod
    def set_log_file_max_lines(log_file, max_lines):
        if os.path.isfile(log_file):
            # Read the contents of the log file
            with open(log_file, "r", encoding="utf-8") as fileObj:
                lines = fileObj.readlines()
            # Count the number of lines in the file
            num_lines = len(lines)
            if num_lines > max_lines:
                # Calculate the number of lines to be deleted
                num_lines_to_delete = num_lines - max_lines
                if num_lines_to_delete > 0:
                    # Open the log file in write mode and truncate it
                    with open(log_file, "w", encoding="utf-8") as fileObj:
                        # Write the remaining lines back to the log file
                        fileObj.writelines(lines[num_lines_to_delete:])
                filename = os.path.basename(log_file)
                HealthCheck.print3(f"Number of old lines deleted from log file '{filename}': {num_lines_to_delete}")

    @staticmethod
    def saveConfig():
        with open(os.path.join(config.packageFolder, "config.py"), "w", encoding="utf-8") as fileObj:
            for name in dir(config):
                excludeConfigList = [
                    "selectAll",
                    "open",
                    "read_suggestions",
                    "currentVerses",
                    "mainFile",
                    "packageFolder",
                    "thisPlatform",
                    "divider",
                    "pipIsUpdated",
                    "isPygameInstalled",
                    "clipboard",
                    "multilineInput",
                    "promptStyle1",
                    "promptStyle2",
                    "terminalColors",
                    "addPathAt",
                    "defaultEntry",
                    "predefinedContextTemp",
                    "pagerContent",
                    "new_chat_response",
                    "currentMessages",
                    "tempChunk",
                ]
                excludeConfigList = excludeConfigList + config.excludeConfigList
                if not name.startswith("__") and not name in excludeConfigList:
                    try:
                        value = eval(f"config.{name}")
                        if not callable(value) and not str(value).startswith("<"):
                            fileObj.write("{0} = {1}\n".format(name, pprint.pformat(value)))
                    except:
                        pass

    @staticmethod
    def restartApp():
        HealthCheck.print2(f"Restarting Search Bible AI ...")
        os.system(f"{sys.executable} {config.mainFile}")
        exit(0)

    @staticmethod
    def isPackageInstalled(package):
        return True if shutil.which(package.split(" ", 1)[0]) else False

    @staticmethod
    def getFiles():
        # option 1: config.storagedirectory; user custom folder
        if not hasattr(config, "storagedirectory") or (config.storagedirectory and not os.path.isdir(config.storagedirectory)):
            config.storagedirectory = ""
        if config.storagedirectory:
            return config.storagedirectory
        # option 2: defaultStorageDir; located in user home directory
        defaultStorageDir = os.path.join(os.path.expanduser('~'), "searchbible")
        try:
            Path(defaultStorageDir).mkdir(parents=True, exist_ok=True)
        except:
            pass
        if os.path.isdir(defaultStorageDir):
            return defaultStorageDir
        # option 3: directory "files" in app directory; to be deleted on every upgrade
        else:
            return os.path.join(config.packageFolder, "files")

    @staticmethod
    def getCliOutput(cli):
        try:
            process = subprocess.Popen(cli, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, *_ = process.communicate()
            return stdout.decode("utf-8")
        except:
            return ""

    def getClipboardText():
        if not hasattr(config, "clipboard"):
            config.clipboard = PyperclipClipboard()
        return HealthCheck.getCliOutput("termux-clipboard-get") if config.terminalEnableTermuxAPI else config.clipboard.get_data().text

    @staticmethod
    def spinning_animation(stop_event):
        while not stop_event.is_set():
            for symbol in "|/-\\":
                print(symbol, end="\r")
                time.sleep(0.1)
        #print("\r", end="")
        #print(" ", end="")

    @staticmethod
    def startSpinning():
        config.stop_event = threading.Event()
        config.spinner_thread = threading.Thread(target=HealthCheck.spinning_animation, args=(config.stop_event,))
        config.spinner_thread.start()

    @staticmethod
    def stopSpinning():
        try:
            config.stop_event.set()
            config.spinner_thread.join()
        except:
            pass

    @staticmethod
    def simplePrompt(inputIndicator="", validator=None, default="", accept_default=False, completer=None, promptSession=None, style=None, is_password=False, bottom_toolbar=None, custom_key_bindings=None):
        this_key_bindings = KeyBindings()

        @this_key_bindings.add("c-q")
        def _(event):
            buffer = event.app.current_buffer
            buffer.text = config.exit_entry
            buffer.validate_and_handle()

        available_key_bindings = [
            this_key_bindings,
            prompt_shared_key_bindings,
        ]
        if custom_key_bindings:
            available_key_bindings.append(custom_key_bindings)
        this_key_bindings = merge_key_bindings(available_key_bindings)

        config.selectAll = False
        inputPrompt = promptSession.prompt if promptSession is not None else prompt
        if not hasattr(config, "clipboard"):
            config.clipboard = PyperclipClipboard()
        if not inputIndicator:
            inputIndicator = [
                ("class:indicator", ">>> "),
            ]
        userInput = inputPrompt(
            inputIndicator,
            key_bindings=this_key_bindings,
            bottom_toolbar=bottom_toolbar if bottom_toolbar is not None else f" [ctrl+q] {config.exit_entry}",
            #enable_system_prompt=True,
            swap_light_and_dark_colors=Condition(lambda: not config.terminalResourceLinkColor.startswith("ansibright")),
            style=style,
            validator=validator,
            #multiline=Condition(lambda: config.multilineInput),
            default=default,
            accept_default=accept_default,
            completer=completer,
            is_password=is_password,
            mouse_support=Condition(lambda: config.mouseSupport),
            clipboard=config.clipboard,
        )
        userInput = textwrap.dedent(userInput) # dedent to work with code block
        return userInput if hasattr(config, "addPathAt") and config.addPathAt else userInput.strip()

    @staticmethod
    def getStringWidth(text):
        width = 0
        for character in text:
            width += wcwidth.wcwidth(character)
        return width

    @staticmethod
    def getPygmentsStyle():
        theme = config.pygments_style if hasattr(config, "pygments_style") and config.pygments_style else "stata-dark" if not config.terminalResourceLinkColor.startswith("ansibright") else "stata-light"
        return style_from_pygments_cls(get_style_by_name(theme))

    @staticmethod
    def print(content):
        print_formatted_text(HTML(content))

    @staticmethod
    def print2(content):
        print_formatted_text(HTML(f"<{config.terminalPromptIndicatorColor2}>{content}</{config.terminalPromptIndicatorColor2}>"))

    @staticmethod
    def print3(content):
        splittedContent = content.split(": ", 1)
        if len(splittedContent) == 2:
            key, value = splittedContent
            print_formatted_text(HTML(f"<{config.terminalPromptIndicatorColor2}>{key}:</{config.terminalPromptIndicatorColor2}> {value}"))
        else:
            config.print2(splittedContent)

    @staticmethod
    def print4(content):
        splittedContent = content.split(") ", 1)
        if len(splittedContent) == 2:
            key, value = splittedContent
            print_formatted_text(HTML(f"<{config.terminalPromptIndicatorColor2}>{key})</{config.terminalPromptIndicatorColor2}> {value}"))
        else:
            config.print2(splittedContent)

    @staticmethod
    def getEmbeddingFunction(embeddingModel=None):
        # import statement is placed here to make this file compatible on Android
        from chromadb.utils import embedding_functions
        embeddingModel = embeddingModel if embeddingModel is not None else config.embeddingModel
        if embeddingModel == "text-embedding-ada-002":
            return embedding_functions.OpenAIEmbeddingFunction(api_key=config.openaiApiKey, model_name="text-embedding-ada-002")
        return embedding_functions.SentenceTransformerEmbeddingFunction(model_name=embeddingModel) # support custom Sentence Transformer Embedding models by modifying config.embeddingModel

    @staticmethod
    def changeAPIkey():
        print("Enter your OpenAI API Key [required]:")
        apikey = prompt(default=config.openaiApiKey, is_password=True)
        if apikey and not apikey.strip().lower() in (config.cancel_entry, config.exit_entry):
            config.openaiApiKey = apikey
        #HealthCheck.checkCompletion()

    @staticmethod
    @check_openai_errors
    def checkCompletion():
        # instantiate a client that can shared with plugins
        os.environ["OPENAI_API_KEY"] = config.openaiApiKey
        client = OpenAI()
        client.chat.completions.create(
            model="gpt-3.5-turbo",
            messages=[{"role": "user", "content" : "hello"}],
            n=1,
            max_tokens=10,
        )
        # set variable 'OAI_CONFIG_LIST' to work with pyautogen
        oai_config_list = []
        for model in HealthCheck.models:
            oai_config_list.append({"model": model, "api_key": config.openaiApiKey})
        os.environ["OAI_CONFIG_LIST"] = json.dumps(oai_config_list)

    # The following method was modified from source:
    # https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
    @staticmethod
    def count_tokens_from_messages(messages, model=""):
        if not model:
            model = config.chatGPTApiModel

        """Return the number of tokens used by a list of messages."""
        try:
            encoding = tiktoken.encoding_for_model(model)
        except KeyError:
            print("Warning: model not found. Using cl100k_base encoding.")
            encoding = tiktoken.get_encoding("cl100k_base")
        if model in {
                "gpt-3.5-turbo",
                "gpt-3.5-turbo-0613",
                "gpt-3.5-turbo-16k",
                "gpt-3.5-turbo-16k-0613",
                "gpt-4-1106-preview",
                "gpt-4-0314",
                "gpt-4-32k-0314",
                "gpt-4",
                "gpt-4-0613",
                "gpt-4-32k",
                "gpt-4-32k-0613",
            }:
            tokens_per_message = 3
            tokens_per_name = 1
        elif model == "gpt-3.5-turbo-0301":
            tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            tokens_per_name = -1  # if there's a name, the role is omitted
        elif "gpt-3.5-turbo" in model:
            #print("Warning: gpt-3.5-turbo may update over time. Returning num tokens assuming gpt-3.5-turbo-0613.")
            return HealthCheck.count_tokens_from_messages(messages, model="gpt-3.5-turbo-0613")
        elif "gpt-4" in model:
            #print("Warning: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613.")
            return HealthCheck.count_tokens_from_messages(messages, model="gpt-4-0613")
        else:
            raise NotImplementedError(
                f"""count_tokens_from_messages() is not implemented for model {model}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
            )
        num_tokens = 0
        for message in messages:
            num_tokens += tokens_per_message
            if not "content" in message or not message.get("content", ""):
                num_tokens += len(encoding.encode(str(message)))
            else:
                for key, value in message.items():
                    num_tokens += len(encoding.encode(value))
                    if key == "name":
                        num_tokens += tokens_per_name
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens


#!/usr/bin/env python3

import re, glob, os, sys, pprint
#import platform
#if (platform.system() != "Windows"):
from ast import literal_eval

class RegexSearch:

    # A simple search & replace.
    @staticmethod
    def replace(text, searchReplace, multiLine=True):
        for search, replace in searchReplace:
            if multiLine:
                text = re.sub(search, replace, text, flags=re.M)
            else:
                text = re.sub(search, replace, text)
        return text

    # Searching in a loop until a specific pattern is no longer found.
    @staticmethod
    def deepReplace(text, searchPattern, searchReplace, multiLine=True):
        p = re.compile(searchPattern, flags=re.M) if multiLine else re.compile(searchPattern)
        while p.search(text):
            for search, replace in searchReplace:
                if multiLine:
                    text = re.sub(search, replace, text, flags=re.M)
                else:
                    text = re.sub(search, replace, text)
        return text

    def searchFile(self, inputFile, searchReplace=None, overwrite=False, deepSearchPattern=None):
        # set output filename
        path, filename = os.path.split(inputFile)
        if overwrite:
            outputFilename = filename
            outputFile = inputFile
        else:
            outputFilename = "replaced_{0}".format(filename)
            outputFile = os.path.join(path, outputFilename)
        # open file and read input text
        print("Processing '{0}' ...".format(filename))
        with open(inputFile, "r", encoding="utf-8") as f:
            newData = f.read()
        # work on non-empty text
        if newData:
            # search and replace the text
            newData = self.processInputText(newData, searchReplace, deepSearchPattern)
            # save output text in a separate file
            with open(outputFile, "w", encoding="utf-8") as f:
                f.write(newData)
                print("Processed.  Saved output as '{0}'.".format(outputFilename))
        else:
            print("No data is read.")

    def searchFilesInFolder(self, folder, searchReplace=None, overwrite=False, deepSearchPattern=None):
        fileList = glob.glob(folder+"/*")
        for file in fileList:
            if os.path.isfile(file):
                self.searchFile(file, searchReplace, overwrite, deepSearchPattern)

    def processInput(self, inputName, searchReplace=None, overwrite=False, deepSearchPattern=None):
        # check if user's input is a file or a folder
        if os.path.isfile(inputName):
            self.searchFile(inputName, searchReplace, overwrite, deepSearchPattern)
        elif os.path.isdir(inputName):
            self.searchFilesInFolder(inputName, searchReplace, overwrite, deepSearchPattern)
        else:
            print("\""+inputName+"\"", "is not found!")

    def processInputText(self, text, searchReplace=None, deepSearchPattern=None):
        if searchReplace:
            if deepSearchPattern:
                text = self.deepReplace(text, deepSearchPattern, searchReplace)
            else:
                text = self.replace(text, searchReplace)
        # Advanced users add customised search & items below:
        # an example of a simple search & replace
#        searchReplace = (
#            ('search1', 'replace1'),
#            ('search2', 'replace2'),
#            ('search3', 'replace3'),
#        )
#        text = self.replace(text, searchReplace)
        # an example of a repetitive search until a pre-defined pattern is no longer found.
#        searchPattern = 'pattern'
#        searchReplace = (
#            ('search1', 'replace1'),
#            ('search2', 'replace2'),
#            ('search3', 'replace3'),
#        )
#        text = self.deepReplace(text, searchPattern, searchReplace)
        return text

if __name__ == '__main__':
    # sys.platform
    # https://docs.python.org/3/library/sys.html#sys.platform
    if (sys.platform == "linux") or (sys.platform == "darwin"):
        import readline

    inputName = ""
    arguments = sys.argv

    # Usage:
    # RegexSearch.py [a nested tuple of search & replace items] [file1 / folder1] [file2 / folder2] [file3 / folder3] ...
    # Support multple files / folders, starting from the second parameter.
    # To make this script executable, run:
    # chmod +x RegexSearch.py
    # To provide a nested tuple / list of search & replace items on command line:
    # e.g. ./RegexSearch.py '(("test", "TEST"), ("[a-e]", r"x\1x"))' test.txt test1.txt
    # To work with dynamic scripting, edit the content of function "processInputText" and run:
    # e.g. ./RegexSearch.py '()' test.txt test1.txt
    # The first parameter is optional to work on a single file or a single folder.  This assumes power users edit the content of function "processInputText" according to their own needs.
    # e.g. ./RegexSearch.py test.txt
    # To run interactive mode, simply run:
    # ./RegexSearch.py
    if len(arguments) > 2:
        try:
            searchReplace = literal_eval(arguments[1])
        except:
            print("Process cancelled!  Reason: Search & replace items are not properly formatted.")
            exit()
        if searchReplace and isinstance(searchReplace, (list, tuple)):
            regexSearch = RegexSearch()
            overwrite = (arguments[2] == "--overwrite")
            if overwrite and (len(arguments) > 3):
                files = arguments[3:]
            else:
                files = arguments[2:]
            if files:
                for inputName in files:
                    if inputName:
                        try:
                            regexSearch.processInput(inputName, searchReplace=searchReplace, overwrite=overwrite)
                        except:
                            print("Process cancelled!  Reason: Search & replace items are not properly formatted.")
                            exit()
    elif len(arguments) == 2 and not arguments[1] == "--deep":
        inputName = arguments[1]
        if inputName:
            RegexSearch().processInput(inputName)
    else:
        # User Interaction
        print("[A search & seplace utility, written by Eliran Wong]")
        # Ask for name(s) of file(s) / folder(s)
        checkMultiple = input("Work on more than one file or folder? [yes/No] ").lower()
        multiple = (checkMultiple == "yes" or checkMultiple == "y")
        if multiple:
            print("Enter the names of your files / folders, separated by comma ',', for examples:")
            print("filename1.txt, filename2.txt, folderName1, folderName2")
            inputNames = input("Enter here: ")
            inputNames = [inputName.strip() for inputName in inputNames.split(",")]
            print("Total of {0} item(s) entered: {1}".format(len(inputNames), inputNames))
        else:
            inputName = input("Enter a file or folder: ")
        # Ask if overwriting original file(s)
        checkOverwrite = input("Overwrite? [yes/No] ").lower()
        overwrite = (checkOverwrite == "yes" or checkOverwrite == "y")
        # check if running deep mode
        deepSearchPattern = None
        if len(arguments) == 2 and arguments[1] == "--deep":
            print("Deep mode is enabled.  Search & replace are performed repetitively until a pre-defined pattern is no longer found.")
            userPattern = input("Define a pattern here: ")
            checkPatternLiteral = input("Is this pattern a string literal? [yes/No] ").lower()
            patternLiteral = (checkPatternLiteral == "yes" or checkPatternLiteral == "y")
            if userPattern:
                if patternLiteral:
                    try:
                        deepSearchPattern = literal_eval(userPattern)
                    except:
                        print("Deep mode is disabled!  Reason: Entered pattern is not properly formatted.")
                        deepSearchPattern = None
                else:
                    deepSearchPattern = userPattern
            else:
                print("Deep mode is disabled!  Reason: Entered pattern is empty.")
        # formulate a nested list of tuples, according to user input
        searchReplace = []
        addSearchReplace = True
        while addSearchReplace:
            print("Adding a pair of search & replace items ...")
            search = input("Search: ")
            if search:
                replace = input("Replace: ")
                checkLiteral = input("Are these items string literals? [yes/No] ").lower()
                literal = (checkLiteral == "yes" or checkLiteral == "y")
                if literal:
                    try:
                        searchReplace.append((literal_eval(search), literal_eval(replace)))
                    except:
                        print("Failed to add this pair of search & replace items!  Reason: One of entered items is not string literal.")
                else:
                    searchReplace.append((search, replace))
                    print("Added successfully.")
            print("Here are your search & replace items: ")
            print(pprint.pformat(searchReplace))
            checkAddSearchReplace = input("Add more search & replace items? [yes/No] ").lower()
            addSearchReplace = (checkAddSearchReplace == "yes" or checkAddSearchReplace == "y")
        try:
            if multiple and inputNames:
                for inputName in inputNames:
                    RegexSearch().processInput(inputName, searchReplace=searchReplace, overwrite=overwrite, deepSearchPattern=deepSearchPattern)
            elif not multiple and inputName:
                RegexSearch().processInput(inputName, searchReplace=searchReplace, overwrite=overwrite, deepSearchPattern=deepSearchPattern)
        except:
            print("Process cancelled!  Reason: One of search & replace items is not properly formatted.")
            exit()



from searchbible import config
import re, html, base64, os
#from util.BibleVerseParser import BibleVerseParser
import urllib.parse
try:
    import html_text
    isHtmlTextInstalled = True
except:
    isHtmlTextInstalled = False

try:
    from bs4 import BeautifulSoup
    import html5lib
    isBeautifulsoup4Installed = True
except:
    isBeautifulsoup4Installed = False

class TextUtil:

    @staticmethod
    def formatConfigLabel(text):
        text = re.sub("([a-z])([A-Z])", r"\1 \2", text)
        words = text.split(" ")
        words[0] = words[0].capitalize()
        return " ".join(words)

    @staticmethod
    def getQueryPrefix():
        return "PRAGMA case_sensitive_like = {0}; ".format("true" if config.enableCaseSensitiveSearch else "false")

    @staticmethod
    def regexp(expr, item):
        reg = re.compile(expr, flags=0 if config.enableCaseSensitiveSearch else re.IGNORECASE)
        #return reg.match(item) is not None
        return reg.search(item) is not None

    @staticmethod
    def highlightSearchString(text, searchString):
        if searchString == "z":
            return text
        if config.enableCaseSensitiveSearch:
            return re.sub("({0})".format(searchString), r"<z>\1</z>", text, flags=0)
        else:
            return re.sub("({0})".format(searchString), r"<z>\1</z>", text, flags=re.IGNORECASE)

    @staticmethod
    # a fallback method when Prompt_toolkit formatted text does not work
    def convertHtmlTagToColorama(text):
        """"
        # prompt-toolkit ansi color names
        config.terminalColors = {
            "ansidefault": "ansidefault",
            "ansiblack": "ansiwhite",
            "ansired": "ansibrightred",
            "ansigreen": "ansibrightgreen",
            "ansiyellow": "ansibrightyellow",
            "ansiblue": "ansibrightblue",
            "ansimagenta": "ansibrightmagenta",
            "ansicyan": "ansibrightcyan",
            "ansigray": "ansibrightblack",
            "ansiwhite": "ansiblack",
            "ansibrightred": "ansired",
            "ansibrightgreen": "ansigreen",
            "ansibrightyellow": "ansiyellow",
            "ansibrightblue": "ansiblue",
            "ansibrightmagenta": "ansimagenta",
            "ansibrightcyan": "ansicyan",
            "ansibrightblack": "ansigray",
        }
        """
        # Reference: https://github.com/tartley/colorama/blob/master/colorama/ansi.py
        # standard colours: "RESET", "BLACK", "WHITE", "RED", "GREEN", "YELLOW", "BLUE", "MAGENTA", "CYAN"
        # extended colours: "LIGHTBLACK_EX", "LIGHTRED_EX", "LIGHTGREEN_EX", "LIGHTYELLOW_EX", "LIGHTBLUE_EX", "LIGHTMAGENTA_EX", "LIGHTCYAN_EX", "LIGHTWHITE_EX"
        try:
            from colorama import Fore, Back, Style

            searchReplace = (
                ("<b>", "\033[1m"),
                ("<i>", "\033[3m"),
                ("<u>", "\033[4m"),
                ("</b>", "\033[0m"),
                ("</i>", "\033[0m"),
                ("</u>", "\033[0m"),
            )
            for search, replace in searchReplace:
                text = text.replace(search, replace)

            searchReplace = (
                ("</ansi[^<>]+?>", Fore.RESET),
                ("</tmvs>|</tmsh>", Style.RESET_ALL),
                ("""<tm[a-z][a-z] fg="([^<>]*?)" bg="([^<>]*?)">""", r"<\1><BG.\2>"),

                ("<ansidefault>", Fore.RESET),

                ("<ansiblack>", Fore.BLACK),
                ("<ansired>", Fore.RED),
                ("<ansigreen>", Fore.GREEN),
                ("<ansiyellow>", Fore.YELLOW),
                ("<ansiblue>", Fore.BLUE),
                ("<ansimagenta>", Fore.MAGENTA),
                ("<ansicyan>", Fore.CYAN),
                ("<ansigray>", Fore.WHITE),
                ("<ansiwhite>", Fore.WHITE),

                ("<ansibrightblack>", Fore.LIGHTBLACK_EX),
                ("<ansibrightred>", Fore.LIGHTRED_EX),
                ("<ansibrightgreen>", Fore.LIGHTGREEN_EX),
                ("<ansibrightyellow>", Fore.LIGHTYELLOW_EX),
                ("<ansibrightblue>", Fore.LIGHTBLUE_EX),
                ("<ansibrightmagenta>", Fore.LIGHTMAGENTA_EX),
                ("<ansibrightcyan>", Fore.LIGHTCYAN_EX),

                ("<BG\.ansidefault>", Back.RESET),

                ("<BG\.ansiblack>", Back.BLACK),
                ("<BG\.ansired>", Back.RED),
                ("<BG\.ansigreen>", Back.GREEN),
                ("<BG\.ansiyellow>", Back.YELLOW),
                ("<BG\.ansiblue>", Back.BLUE),
                ("<BG\.ansimagenta>", Back.MAGENTA),
                ("<BG\.ansicyan>", Back.CYAN),
                ("<BG\.ansigray>", Back.WHITE),
                ("<BG\.ansiwhite>", Back.WHITE),

                ("<BG\.ansibrightblack>", Back.LIGHTBLACK_EX),
                ("<BG\.ansibrightred>", Back.LIGHTRED_EX),
                ("<BG\.ansibrightgreen>", Back.LIGHTGREEN_EX),
                ("<BG\.ansibrightyellow>", Back.LIGHTYELLOW_EX),
                ("<BG\.ansibrightblue>", Back.LIGHTBLUE_EX),
                ("<BG\.ansibrightmagenta>", Back.LIGHTMAGENTA_EX),
                ("<BG\.ansibrightcyan>", Back.LIGHTCYAN_EX),

                ("<[^<>]*?>", ""),
            )
            for search, replace in searchReplace:
                text = re.sub(search, replace, text)
        except:
            text = re.sub("<[^<>]*?>", "", text)

        return text

    @staticmethod
    def colourTerminalText(text):
        searchReplace = (
            #("(<u><b>|<h>|<h[0-9]>|<h[0-9] .*?>)", r"「{0}」".format(config.terminalHeadingTextColor)),
            #("(</h[0-9]+?>|</h>|</b></u>)", r"「/{0}」".format(config.terminalHeadingTextColor)),
            # make sure tags are paired
            ("(<u><b>)(.*?)(</b></u>)", r"\1「{0}」\2「/{0}」\3".format(config.terminalHeadingTextColor)),
            ("(<h>|<h[0-9]>|<h[0-9] .*?>)(.*?)(</h[0-9]+?>|</h>)", r"\1「{0}」\2「/{0}」\3".format(config.terminalHeadingTextColor)),
            #("(<ref>|<ref .*?>|<grk>|<grk .*?>|<heb>|<heb .*?>)", r"「{0}」".format(config.terminalResourceLinkColor)),
            #("(</heb>|</grk>|</ref>)", r"「/{0}」".format(config.terminalResourceLinkColor)),
            ("(<ref>|<ref .*?>)(.*?)(</ref>)", r"\1「{0}」\2「/{0}」\3".format(config.terminalResourceLinkColor)),
            ("(<heb>|<heb .*?>)(.*?)(</heb>)", r"\1「b」「{0}」\2「/{0}」「/b」\3".format(config.terminalResourceLinkColor)),
            ("(<grk>|<grk .*?>)(.*?)(</grk>)", r"\1「{0}」\2「/{0}」\3".format(config.terminalResourceLinkColor)),
            #("(<vid .*?>)", r"「{0}」".format(config.terminalVerseNumberColor)),
            #("(</vid>)", r"「/{0}」".format(config.terminalVerseNumberColor)),
            ("(<vid>|<vid .*?>)(.*?)(</vid>)", r"\1「{0}」\2「/{0}」\3".format(config.terminalResourceLinkColor)),
            #("<z>", """「tmsh fg="{1}" bg="{0}"」""".format(config.terminalSearchHighlightBackground, config.terminalSearchHighlightForeground)),
            #("</z>", "「/tmsh」"),
            ("(<z>)(.*?)(</z>)", r"""「tmsh fg="{1}" bg="{0}"」\2「/tmsh」""".format(config.terminalSearchHighlightBackground, config.terminalSearchHighlightForeground)),
            # basic html
            ("<(b|u|i)>", r"「\1」"),
            ("</(b|u|i)>", r"「/\1」"),
        )
        for search, replace in searchReplace:
            text = re.sub(search, replace, text)
        return text

    @staticmethod
    def htmlToPlainText(content, colours=True):
        content = content.replace("<hr>", "<br><b>--------------------</b><br>")
        if config.runMode == "terminal":
            content = re.sub(r"""<ref onclick="lex\('(H[0-9]+?)'\)" class="G\1" onmouseover="ld\('\1'\); hl1\('','','\1'\)" onmouseout="hl0\('','','\1'\)">\1</ref>""", r"[<ref>\1</ref> ] ", content)
            content = re.sub("""(<heb|<grk)( [^<>]*?onclick="luW\([0-9]+?,')([0-9]+?)('[^<>]*?>)""", r"[<ref>\3</ref> ]\1\2\3\4", content)
            content = re.sub("""(<ref onclick="[^<>]+?\(')([^<>]+?)('\)">)""", r"[<ref>\2</ref> ] ", content)
        # Format text colours
        if config.runMode == "terminal" and colours:
            content = TextUtil.colourTerminalText(content)
        # cconvert text
        if isHtmlTextInstalled:
            content = html_text.extract_text(content)
        elif isBeautifulsoup4Installed:
            content = re.sub("(</th>|</td>)", r"\1&emsp;", content)
            content = re.sub("(<br>|<br/>|</tr>)", r"\1\n", content)
            content = re.sub("(</h[0-9]>|</p>|</div>|<hr>)", r"\1\n\n", content)
            content = BeautifulSoup(content, "html5lib").get_text()
        else:
            content = re.sub("<br/?>|<br>", "\n", content)
            content = re.sub('<[^<]+?>', '', content)
        if config.runMode == "terminal" and not colours:
            content = re.sub("""「ansi[^「」]+？」|「tm[a-z][a-z] fg="[^「」]*?" bg="[^「」]*?"」|「/tm[a-z][a-z]」""", "", content)
        elif config.runMode == "terminal" and colours:
            searchReplace = (
                ("「ansiblack」", "<ansiblack>"),
                ("「ansired」", "<ansired>"),
                ("「ansigreen」", "<ansigreen>"),
                ("「ansiyellow」", "<ansiyellow>"),
                ("「ansiblue」", "<ansiblue>"),
                ("「ansimagenta」", "<ansimagenta>"),
                ("「ansicyan」", "<ansicyan>"),
                ("「ansigray」", "<ansigray>"),
                ("「ansiwhite」", "<ansiwhite>"),
                ("「ansibrightred」", "<ansibrightred>"),
                ("「ansibrightgreen」", "<ansibrightgreen>"),
                ("「ansibrightyellow」", "<ansibrightyellow>"),
                ("「ansibrightblue」", "<ansibrightblue>"),
                ("「ansibrightmagenta」", "<ansibrightmagenta>"),
                ("「ansibrightcyan」", "<ansibrightcyan>"),
                ("「ansibrightblack」", "<ansibrightblack>"),
                ("「/ansiblack」", "</ansiblack>"),
                ("「/ansired」", "</ansired>"),
                ("「/ansigreen」", "</ansigreen>"),
                ("「/ansiyellow」", "</ansiyellow>"),
                ("「/ansiblue」", "</ansiblue>"),
                ("「/ansimagenta」", "</ansimagenta>"),
                ("「/ansicyan」", "</ansicyan>"),
                ("「/ansigray」", "</ansigray>"),
                ("「/ansiwhite」", "</ansiwhite>"),
                ("「/ansibrightred」", "</ansibrightred>"),
                ("「/ansibrightgreen」", "</ansibrightgreen>"),
                ("「/ansibrightyellow」", "</ansibrightyellow>"),
                ("「/ansibrightblue」", "</ansibrightblue>"),
                ("「/ansibrightmagenta」", "</ansibrightmagenta>"),
                ("「/ansibrightcyan」", "</ansibrightcyan>"),
                ("「/ansibrightblack」", "</ansibrightblack>"),
                ("""「(tm[a-z][a-z] fg="[^「」]*?" bg="[^「」]*?")」""", r"<\1>"),
                ("「(/tm[a-z][a-z])」", r"<\1>"),
                ("「(b|u|i)」", r"<\1>"),
                ("「/(b|u|i)」", r"</\1>"),
                #("[ ]*「Fore.RESET」", Fore.RESET),
                #("[ ]*「Back.RESET」", Back.RESET),
                #("[ ]*「Style.RESET_ALL」", Style.RESET_ALL),
            )
            for search, replace in searchReplace:
                content = re.sub(search, replace, content)

        searchReplace = (
            (" audiotrack", ""),
            (" [ ]+?([^ ])", r" \1"),
        )
        for search, replace in searchReplace:
            content = re.sub(search, replace, content)
        # fine tune
        if config.runMode == "terminal":
            content = re.sub(r"""(G[0-9]+?)'\)" class="G\1" onmouseover="ld\('\1'\); hl1\('','','\1'\)" onmouseout="hl0\('','','\1""", r"\1", content)
        return content

    @staticmethod
    def imageToText(filepath):
        fileBasename = os.path.basename(filepath)
        *_, fileExtension = os.path.splitext(fileBasename)
        if fileExtension.lower() in (".png", ".jpg", ".jpeg", ".bmp", ".gif"):
            # read a binary file
            with open(filepath, "rb") as fileObject:
                binaryData = fileObject.read()
                encodedData = base64.b64encode(binaryData)
                binaryString = encodedData.decode("ascii")
                htmlTag = '<img src="data:image/{2};base64,{0}" alt="{1}">'.format(binaryString, fileBasename, fileExtension[1:])
            return htmlTag
        else:
            return "[File type not supported!]"

    @staticmethod
    def formulateUBACommandHyperlink(text):
        # Create hyperlink to UBA command
        # work on text formatted like ***[CROSSREFERENCE:::John 3:16@An hyperlink link to open cross-references of John 3:16]
        return re.sub("\*\*\*\[([^'{0}]*?)@([^'{0}]*?)\]".format('"\*\[\]@'), r"<ref onclick={0}document.title='\1'{0}>\2</ref>".format('"'), text)

    @staticmethod
    def fixTextHighlighting(text):
        # fix searching LXX / SBLGNT words
        text = re.sub(r"<z>([LS][0-9]+?)</z>'\)"'"'">(.*?)</grk>", r"\1'\)"'"'r"><z>\2</z></grk>", text)
        # remove misplacement of tags <z> & </z>
        p = re.compile("(<[^<>]*?)<z>(.*?)</z>", flags=re.M)
        s = p.search(text)
        while s:
            text = re.sub(p, r"\1\2", text)
            s = p.search(text)
        return text

    @staticmethod
    def plainTextToUrl(text):
        # https://wiki.python.org/moin/EscapingHtml
        text = html.escape(text)
        searchReplace = (
            (" ", "%20"),
            ("\n", "%0D%0A"),
        )
        for search, replace in searchReplace:
            text = text.replace(search, replace)
        return text

    # Return digits from a string
    @staticmethod
    def getDigits(text):
        return ''.join(c for c in text if c.isdigit())

    # Generate a web link for sharing
    @staticmethod
    def getWeblink(command, server=""):
        # https://stackoverflow.com/questions/40557606/how-to-url-encode-in-python-3
        command = urllib.parse.quote(command)
        htmlPages = {
            "ENG": "index.html",
            "TC": "traditional.html",
            "SC": "simplified.html",
        }
        htmlPage = "" if config.webUBAServer.endswith(".html") and not server else "/{0}".format(htmlPages.get(htmlPages[config.standardAbbreviation], "index.html"))
        return "{0}{1}?cmd={2}".format(server if server else config.webUBAServer, htmlPage, command)

    # Remove Hebrew vowels or Greek accents
    @staticmethod
    def removeVowelAccent(text):
        searchReplace = (
            (r"[\֑\֒\֓\֔\֕\֖\֗\֘\֙\֚\֛\֜\֝\֞\֟\֠\֡\֣\֤\֥\֦\֧\֨\֩\֪\֫\֬\֭\֮\ֽ\ׄ\ׅ\‍\‪\‬\̣\ְ\ֱ\ֲ\ֳ\ִ\ֵ\ֶ\ַ\ָ\ֹ\ֺ\ֻ\ׂ\ׁ\ּ\ֿ\־\׀\׆]", ""),
            ("[שׂשׁ]", "ש"),
            ("[ἀἄᾄἂἆἁἅᾅἃάᾴὰᾶᾷᾳ]", "α"),
            ("[ἈἌἎἉἍἋ]", "Α"),
            ("[ἐἔἑἕἓέὲ]", "ε"),
            ("[ἘἜἙἝἛ]", "Ε"),
            ("[ἠἤᾔἢἦᾖᾐἡἥἣἧᾗᾑήῄὴῆῇῃ]", "η"),
            ("[ἨἬἪἮἩἭἫ]", "Η"),
            ("[ἰἴἶἱἵἳἷίὶῖϊΐῒ]", "ι"),
            ("[ἸἼἹἽ]", "Ι"),
            ("[ὀὄὂὁὅὃόὸ]", "ο"),
            ("[ὈὌὉὍὋ]", "Ο"),
            ("[ῥ]", "ρ"),
            ("[Ῥ]", "Ρ"),
            ("[ὐὔὒὖὑὕὓὗύὺῦϋΰῢ]", "υ"),
            ("[ὙὝὟ]", "Υ"),
            ("[ὠὤὢὦᾠὡὥὧᾧώῴὼῶῷῳ]", "ω"),
            ("[ὨὬὪὮὩὭὯ]", "Ω"),
        )
        for search, replace in searchReplace:
            text = re.sub(search, replace, text)
        return text

    # fix note font display
    @staticmethod
    def fixNoteFontDisplay(content):
        if config.overwriteNoteFont:
            content = re.sub("font-family:[^<>]*?([;'{0}])".format('"'), r"font-family:{0}\1".format(config.font),
                             content)
        if config.overwriteNoteFontSize:
            content = re.sub("font-size:[^<>]*?;", "", content)
        return content

    # fix note font display
    @staticmethod
    def fixNoteFont(note):
        note = re.sub("<body style={0}[ ]*?font-family:[ ]*?'[^']*?';[ ]*?font-size:[ ]*?[0-9]+?pt;".format('"'), "<body style={0}font-family:'{1}'; font-size:{2}pt;".format('"', config.font, config.fontSize), note)
        if not config.includeStrictDocTypeInNote:
            note = re.sub("""<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">\n""", "", note)
        return note

    # wrap with html
    @staticmethod
    def htmlWrapper(text, parsing=False, view="study", linebreak=True, html=True):
        searchReplace1 = (
            ("\r\n|\r|\n", "<br>"),
            ("\t", "&emsp;&emsp;"),
        )
        searchReplace2 = (
            ("<br>(<table>|<ol>|<ul>)", r"\1"),
            ("(</table>|</ol>|</ul>)<br>", r"\1"),
            ("<a [^\n<>]*?href=['{0}]([^\n<>]*?)['{0}][^\n<>]*?>".format('"'),
             r"<a href='javascript:void(0)' onclick='website({0}\1{0})'>".format('"')),
            ("onclick='website\({0}([^\n<>]*?).uba{0}\)'".format('"'), r"onclick='uba({0}\1.uba{0})'".format('"'))
        )
        if linebreak:
            for search, replace in searchReplace1:
                text = re.sub(search, replace, text)
        if html:
            for search, replace in searchReplace2:
                text = re.sub(search, replace, text)
        if parsing:
            # Export inline images to external files, so as to improve parsing performance. 
            text = TextUtil.exportAllImages(text)
            text = TextUtil.formulateUBACommandHyperlink(text)
            text = BibleVerseParser(config.parserStandarisation).parseText(text)
        if not "<!DOCTYPE html><html><head><meta charset='utf-8'><title>UniqueBible.app</title>" in text:
            text = TextUtil.wrapHtml(text, view)
        return text

    # export images
    @staticmethod
    def exportAllImages(htmlText):
        config.exportImageNumber = 0
        searchPattern = r'src=(["{0}])data:image/([^<>]+?);[ ]*?base64,[ ]*?([^ <>]+?)\1'.format("'")
        htmlText = re.sub(searchPattern, TextUtil.exportAnImage, htmlText)
        return htmlText

    @staticmethod
    def exportAnImage(match):
        exportFolder = os.path.join("htmlResources", "images", "export")
        if not os.path.isdir(exportFolder):
            os.makedirs(exportFolder)
        quotationMark, ext, asciiString = match.groups()
        # Note the difference between "groups" and "group"
        # wholeString = match.group(0)
        # quotationMark = match.group(1)
        # ext = match.group(2)
        # asciiString = match.group(3)
        config.exportImageNumber += 1
        binaryString = asciiString.encode("ascii")
        binaryData = base64.b64decode(binaryString)
        imageFilename = "tab{0}_image{1}.{2}".format(100, config.exportImageNumber, ext)
        exportPath = os.path.join(exportFolder, imageFilename)
        with open(exportPath, "wb") as fileObject2:
            fileObject2.write(binaryData)
        return "src={0}images/export/{1}{0}".format(quotationMark, imageFilename)

    # wrap with html 2
    @staticmethod
    def wrapHtml(content, view="", book=False):
        fontFamily = config.font
        fontSize = "{0}px".format(config.fontSize)
        if book:
            if config.overwriteBookFontFamily:
                fontFamily = config.overwriteBookFontFamily
            if config.overwriteBookFontSize:
                if type(config.overwriteBookFontSize) == str:
                    fontSize = config.overwriteBookFontSize
                elif type(config.overwriteBookFontSize) == int:
                    fontSize = "{0}px".format(config.overwriteBookFontSize)
        bcv = (config.studyText, config.studyB, config.studyC, config.studyV) if view == "study" else (config.mainText, config.mainB, config.mainC, config.mainV)
        activeBCVsettings = "<script>var activeText = '{0}'; var activeB = {1}; var activeC = {2}; var activeV = {3};</script>".format(*bcv)
        html = ("""<!DOCTYPE html><html><head><link rel="icon" href="icons/{9}"><title>UniqueBible.app</title>
                <meta charset="UTF-8">
                <meta name="viewport" content="width=device-width, initial-scale=1.0">
                <meta http-equiv="Cache-Control" content="no-cache, no-store, must-revalidate" />
                <meta http-equiv="Pragma" content="no-cache" />
                <meta http-equiv="Expires" content="0" />"""
                "<style>body {2} font-size: {4}; font-family:'{5}';{3} "
                "zh {2} font-family:'{6}'; {3} "
                ".ubaButton {2} background-color: {10}; color: {11}; border: none; padding: 2px 10px; text-align: center; text-decoration: none; display: inline-block; font-size: 17px; margin: 2px 2px; cursor: pointer; {3}"
                "{8}</style>"
                "<link id='theme_stylesheet' rel='stylesheet' type='text/css' href='css/{7}.css?v=1.065'>"
                "<link id='theme_stylesheet' rel='stylesheet' type='text/css' href='css/custom.css?v=1.065'>"
                "<script src='js/common.js?v=1.065'></script>"
                "<script src='js/{7}.js?v=1.065'></script>"
                "<script src='w3.js?v=1.065'></script>"
                "<script src='js/http_server.js?v=1.065'></script>"
                """<script>
                var target = document.querySelector('title');
                var observer = new MutationObserver(function(mutations) {2}
                    mutations.forEach(function(mutation) {2}
                        ubaCommandChanged(document.title);
                    {3});
                {3});
                var config = {2}
                    childList: true,
                {3};
                observer.observe(target, config);
                </script>"""
                "{0}"
                """<script>var versionList = []; var compareList = []; var parallelList = [];
                var diffList = []; var searchList = [];</script>"""
                "<script src='js/custom.js?v=1.065'></script>"
                "</head><body><span id='v0.0.0'></span>{1}"
                "<p>&nbsp;</p><div id='footer'><span id='lastElement'></span></div><script>loadBible();document.querySelector('body').addEventListener('click', window.parent.closeSideNav);</script></body></html>"
                ).format(activeBCVsettings,
                         content,
                         "{",
                         "}",
                         fontSize,
                         fontFamily,
                         config.fontChinese,
                         config.theme,
                         TextUtil.getHighlightCss(),
                         config.webUBAIcon,
                         config.widgetBackgroundColor,
                         config.widgetForegroundColor,
                         )
        return html

    # get highlight css
    @staticmethod
    def getHighlightCss():
        css = ""
        for i in range(len(config.highlightCollections)):
            code = "hl{0}".format(i + 1)
            css += ".{2} {0} background: {3}; {1} ".format("{", "}", code, config.highlightDarkThemeColours[i] if config.theme == "dark" else config.highlightLightThemeColours[i])
        return css

    # Remove special characters
    @staticmethod
    def removeSpecialCharacters(text):
        searchReplace = (
            (r"[\-\—\,\;\:\\\?\.\·\·\‘\’\‹\›\“\”\«\»\(\)\[\]\{\}\⧼\⧽\〈\〉\*\‿\᾽\⇔\¦]", ""),
        )
        for search, replace in searchReplace:
            text = re.sub(search, replace, text)
        return text


if __name__ == '__main__':

    print(TextUtil.getDigits("abc123def"))



#!/usr/bin/env python

"""
A python utility to parse bible verse references.
It is both useful for tagging and extracting bible verse references from a text.

This was originally created to format resources <a href="https://marvel.bible">https://marvel.bible</a>.
(The original version is available at https://github.com/eliranwong/bible-verse-parser)

This is now integrated into UniqueBible.app as a core utility.

Features:
1. Search for verse references from text file(s)
2. Add taggings on verse references
3. Support books of bible canon and apocrypha
4. Support tagging on chains of refernces, e.g. Rom 1:2, 3, 5, 8; 9:2, 10
5. Support books of one chapter only, like Obadiah 2, Jude 3, 3John 4, etc.
6. Support chapter references [references without verse number specified], e.g. Gen 1, 3-4; 8, 9-10.
7. Support standardisation of book abbreviations and verse reference format.
8. Support parsing multiple files in one go.
9. Support extracting all references in a text.

Usage [standalone]:
Command line: python BibleVerseParser.py

User Interaction:
Prompting question (1) "Enter a file / folder name here: "<br>
Enter the name of a file, which you want to parse.
OR
Enter the name of a directory containing files, which you want you parse.

Prompting question (2) "Do you want to standardise the format of all bible verse references? [YES/NO] "<br>
Enter YES if you want to standardise all verse references with SBL-style-abbreviations and common format like Gen 2:4; Deut 6:4, etc.<br>
Any answers other than "YES" [case-insensitive] skip the standarisation.
"""
#import logging
import os

# File "config.py" is essential for running module "config"
# Create file "config.py" if it is missing.
# The following two lines are written for use of this parser outside UniqueBible.app
import pprint

from searchbible import config

# import modules, which are ESSENTIAL for running BibleVerseParser
import re, glob, sys
from ast import literal_eval
from searchbible.utils.BibleBooks import BibleBooks
from searchbible.utils.RegexSearch import RegexSearch

"""
START - class BibleVerseParser
"""

class BibleVerseParser:

    standardAbbreviation = {}

    # initialisation
    def __init__(self, standardisation, noOfLinesPerChunkForParsing=None):
        #self.logger = logging.getLogger('uba')
        # noOfLinesPerChunkForParsing
        self.noOfLinesPerChunkForParsing = config.noOfLinesPerChunkForParsing if noOfLinesPerChunkForParsing is None else noOfLinesPerChunkForParsing
        # set standard abbreviation, displayed in UniqueBible
        self.updateStandardAbbreviation()
        # set preference of standardisation
        self.standardisation = standardisation
        self.bibleBooksDict = {}
        self.bibleBooksDictCanonBooksOnly = {}
        pattern = re.compile("[a-zA-Z]")
        bookNames = [bookName for bookName in BibleBooks.name2number.keys() if re.search(pattern, bookName)] if config.parseEnglishBooksOnly else BibleBooks.name2number.keys()
        for bookName in bookNames:
            num = int(BibleBooks.name2number[bookName])
            self.bibleBooksDict[bookName] = num
            if num <= 66:
                self.bibleBooksDictCanonBooksOnly[bookName] = num
            if config.useLiteVerseParsing:
                if "." in bookName:
                    bookName = bookName.replace(".", "")
                    self.bibleBooksDict[bookName] = int(num)
                if bookName[0] < 'a':
                    bookName = bookName.lower()
                    self.bibleBooksDict[bookName] = int(num)
        sortedNames = sorted(self.bibleBooksDict.keys())
        self.sortedNames = sorted(sortedNames, key=len, reverse=True)
        sortedNamesCanonBooksOnly = sorted(self.bibleBooksDictCanonBooksOnly.keys())
        self.sortedNamesCanonBooksOnly = sorted(sortedNamesCanonBooksOnly, key=len, reverse=True)

    # function for converting b c v integers to verse reference string
    def bcvToVerseReference(self, b, c, v, *args, isChapter=False):
        bookNo = str(b)
        if bookNo in self.standardAbbreviation:
            abbreviation = self.standardAbbreviation[bookNo]
            if isChapter:
                return "{0} {1}".format(abbreviation, c)
            if args and len(args) == 2:
                c2, v2 = args
                if c2 == c:
                    if v2 > v:
                        return "{0} {1}:{2}-{3}".format(abbreviation, c, v, v2)
                    elif v2 == v:
                        return "{0} {1}:{2}".format(abbreviation, c, v)
                    else:
                        return ""
                elif c2 > c:
                    return "{0} {1}:{2}-{3}:{4}".format(abbreviation, c, v, c2, v2)
            else:
                return "{0} {1}:{2}".format(abbreviation, c, v)
        else:
            return ""

    # update self.standardAbbreviation
    def updateStandardAbbreviation(self):
        self.checkConfig()
        self.standardAbbreviation = BibleBooks().booksMap.get(config.standardAbbreviation, BibleBooks.abbrev["eng"])
        self.standardFullBookName = {key: value[1] for key, value in self.standardAbbreviation.items()}
        self.standardAbbreviation = {key: value[0] for key, value in self.standardAbbreviation.items()}

    # The following two lines are written for use of this parser outside UniqueBible.app
    # This function check if there is an existing value of config.standardAbbreviation
    def checkConfig(self):
        if not hasattr(config, "standardAbbreviation"):
            config.standardAbbreviation = "ENG"
            config.convertChapterVerseDotSeparator = True
            nameValue = (
                ("standardAbbreviation = ", config.standardAbbreviation),
                ("convertChapterVerseDotSeparator = ", config.convertChapterVerseDotSeparator),
            )
            if not os.path.isfile("config.py"):
                with open("config.py", "w", encoding="utf-8") as fileObj:
                    for name, value in nameValue:
                        fileObj.write(name+pprint.pformat(value))

    # To format of all references by using standard abbreviations.
    def standardReference(self, text):
        for booknumber in self.standardAbbreviation:
            abbreviation = self.standardAbbreviation[booknumber]
            searchReplace = (
                ('<ref onclick="bcv\('+booknumber+',([0-9]+?),([0-9]+?)\)">.*?</ref>', '<ref onclick="bcv('+booknumber+r',\1,\2)">'+abbreviation+r' \1:\2</ref>'),
                ('<ref onclick="bcv\('+booknumber+',([0-9]+?),([0-9]+?),([0-9]+?),([0-9]+?)\)">.*?</ref>', '<ref onclick="bcv('+booknumber+r',\1,\2,\3,\4)">'+abbreviation+r' \1:\2-\3:\4</ref>'),
                (r' ([0-9]+?):([0-9]+?)-\1:([0-9]+?)</ref>', r' \1:\2-\3</ref>'),
            )
            text = RegexSearch.replace(text, searchReplace)
        return text

    def parseText(self, text, splitInChunks=True, parseBooklessReferences=False, canonicalOnly=False):
        if splitInChunks:
            parsedText = ""
            chunks = text.splitlines(True)
            chunks = [chunks[x:x+self.noOfLinesPerChunkForParsing] for x in range(0, len(chunks), self.noOfLinesPerChunkForParsing)]
            for chunk in chunks:
                parsedText += self.runParseText("".join(chunk), parseBooklessReferences, canonicalOnly)
        else:
            text = text.replace("\n", "")
            parsedText = ""
            start = 0
            end = 0
            # Do not parse embedded images
            while end < len(text):
                if '<img src="data:image' in text[start:]:
                    imgstart = text.find('<img src="data:image', start)

                    block = text[start:imgstart]
                    parsedText += self.runParseText(block, parseBooklessReferences, canonicalOnly)
                    end = text.find(' />', imgstart+10)+3
                    parsedText += text[imgstart:end]
                    start = end
                else:
                    end = len(text)
                    block = text[start: end]
                    parsedText += self.runParseText(block, parseBooklessReferences, canonicalOnly)
            parsedText = parsedText.replace("</p>", "</p>\n")
        return parsedText

    def runParseText(self, text, parseBooklessReferences=False, canonicalOnly=False):
        # Add a space at the end of the text, to avoid indefinite loop in some of the following processes.
        # This extra space will be removed when parsing is finished.
        text = text + " "

        # remove bcv tags, if any, to avoid duplication of tagging in later steps
        searchPattern = '<ref onclick="bcv\([0-9]+?,[0-9]+?,[0-9][^\(\)]*?\)">(.*?)</ref>'
        searchReplace = (
            (searchPattern, r'\1'),
        )
        text = RegexSearch.deepReplace(text, searchPattern, searchReplace)

        searchBooks = self.sortedNames
        if canonicalOnly:
            searchBooks = self.sortedNamesCanonBooksOnly
        for name in searchBooks:
            # get the string of book name
            bookName = name
            searchReplace = (
                ('\.', r'[\.]*?'), # make dot "." optional for an abbreviation
                ('^([0-9]+?) ', r'\1[ ]*?'), # make space " " optional in some cases
                ('^([I]+?) ', r'\1[ ]*?'),
                ('^(IV) ', r'\1[ ]*?'),
            )
            bookName = RegexSearch.replace(bookName, searchReplace)
            # get assigned book number from dictionary
            bookNumber = str(self.bibleBooksDict[name])
            # search & replace for marking book
            # bypass the rest of parsing if input text contains a single book name only
            if re.match('^('+bookName+') $', text):
                return RegexSearch.replace(text, (('^('+bookName+') $', r'<ref onclick="bcv({0},1,1)">\1 1:1</ref>'.format(bookNumber)),))
            searchReplace = (
                ('('+bookName+') ([0-9])', '『'+bookNumber+r'｜\1』 \2'),
            )
            text = RegexSearch.replace(text, searchReplace)
            if config.parseBookChapterWithoutSpace:
                text = RegexSearch.replace(text, (('('+bookName+')([0-9])', '『'+bookNumber+r'｜\1』 \2'),))

        # In case a dot sign, instead of a colon sign, is used to separate chapter number and verse number.
        if config.convertChapterVerseDotSeparator:
            text = RegexSearch.replace(text, (("』 ([0-9]+?)\.([0-9])", r"』 \1:\2"),))
        searchReplace = (
            # 1st set of taggings
            ('『([0-9]+?)｜([^『』]*?)』 ([0-9]+?):([0-9]+?)([^0-9])', r'<ref onclick="bcv(\1,\3,\4)">\2 \3:\4</ref｝\5'),
            ('『([0-9]+?)｜([^『』]*?)』 ([0-9]+?)([^0-9])', r'<ref onclick="bcv(\1,\3,)">\2 \3</ref｝\4'),
            # fix references without verse numbers
            # fix books with one chapter ONLY; oneChapterBook = [31,57,63,64,65,72,73,75,79,85]
            ('<ref onclick="bcv\((31|57|63|64|65|72|73|75|79|85),([0-9]+?),\)">', r'<ref onclick="bcv(\1,1,\2)">'),
            # fix chapter references without verse number; assign verse number 1 in taggings
            ('<ref onclick="bcv\(([0-9]+?),([0-9]+?),\)">', r'<ref onclick="bcv(\1,\2,1)">＊'),
        )
        text = RegexSearch.replace(text, searchReplace)

        # bypass the rest of parsing if input text contains a single reference only
        if re.match('<ref onclick="bcv[^<>]+?>[^<>]+?</ref｝ ', text):
            text = text.replace("｝", ">")
            return text.replace("＊", "")[:-1]

        # check if tagged references are followed by untagged references, e.g. Book 1:1-2:1; 3:2-4, 5; Jude 1
        if parseBooklessReferences and config.parseBooklessReferences:
            searchReplace = (
                ("[EGH][0-9]+?([^0-9])", r"\1"),
                ("｝[^｝,-–;0-9][^｝<>]*?([0-9])", r"｝; \1"),
                ("([0-9])[^｝,-–;0-9][^｝<>]*?([0-9])", r"\1; \2"),
            )
            text = RegexSearch.replace(text, searchReplace)
        searchPattern = '</ref｝[,-–;][ ]*?[0-9]'
        searchReplace = (
            ('<ref onclick="bcv\(([0-9]+?),([0-9]+?),([0-9]+?)\)">([^｝]*?)</ref｝([,-–;][ ]*?)([0-9]+?):([0-9]+?)([^0-9])', r'<ref onclick="bcv(\1,\2,\3)">\4</ref｝\5<ref onclick="bcv(\1,\6,\7)">\6:\7</ref｝\8'),
            ('<ref onclick="bcv\(([0-9]+?),([0-9]+?),([0-9]+?)\)">([^＊][^｝]*?)</ref｝([,-–;][ ]*?)([0-9]+?)([^:0-9])', r'<ref onclick="bcv(\1,\2,\3)">\4</ref｝\5<ref onclick="bcv(\1,\2,\6)">\6</ref｝\7'),
            ('<ref onclick="bcv\(([0-9]+?),([0-9]+?),([0-9]+?)\)">([^＊][^｝]*?)</ref｝([,-–;][ ]*?)([0-9]+?):([^0-9])', r'<ref onclick="bcv(\1,\2,\3)">\4</ref｝\5<ref onclick="bcv(\1,\2,\6)">\6</ref｝:\7'),
            ('<ref onclick="bcv\(([0-9]+?),([0-9]+?),([0-9]+?)\)">(＊[^｝]*?)</ref｝([,-–;][ ]*?)([0-9]+?)([^:0-9])', r'<ref onclick="bcv(\1,\2,\3)">\4</ref｝\5<ref onclick="bcv(\1,\6,1)">＊\6</ref｝\7'),
            ('<ref onclick="bcv\(([0-9]+?),([0-9]+?),([0-9]+?)\)">(＊[^｝]*?)</ref｝([,-–;][ ]*?)([0-9]+?):([^0-9])', r'<ref onclick="bcv(\1,\2,\3)">\4</ref｝\5<ref onclick="bcv(\1,\6,1)">＊\6</ref｝:\7'),
        )
        text = RegexSearch.deepReplace(text, searchPattern, searchReplace)

        if config.parseClearSpecialCharacters:
            searchReplace = (
                ('『[0-9]+?|([^『』]*?)』', r'\1'),
            )
            text = RegexSearch.replace(text, searchReplace)

        searchReplace = (
            ('(<ref onclick="bcv\([0-9]+?,[0-9]+?,[0-9]+?\)">)＊', r'\1'),
            ('</ref｝', '</ref>'),
        )
        text = RegexSearch.replace(text, searchReplace)

        # handling range of verses
        # e.g. John 3:16 is tagged as <ref onclick="bcv(43,3,16)">John 3:16</ref>
        # e.g. John 3:14-16 is tagged as <ref onclick="bcv(43,3,14,3,16)">John 3:14-16</ref>
        # e.g. John 3:14-4:3 is tagged as <ref onclick="bcv(43,3,14,4,3)">John 3:14-4:3</ref>
        searchPattern = '<ref onclick="bcv\(([0-9]+?),([0-9]+?),([0-9]+?)\)">([^<>]*?)</ref>([-–])<ref onclick="bcv\({0},([0-9]+?),([0-9]+?)\)">'.format(r'\1')
        searchReplace = (
            (searchPattern, r'<ref onclick="bcv(\1,\2,\3,\6,\7)">\4\5'),
        )
        text = RegexSearch.deepReplace(text, searchPattern, searchReplace)

        # final clean up
        searchPattern = '(<ref onclick="bcv[^>]*?)([\(, ])0([1-9])'
        searchReplace = (
            (searchPattern, r'\1\2\3'),
        )
        text = RegexSearch.deepReplace(text, searchPattern, searchReplace)
        text = RegexSearch.deepReplace(text, "<ref [^>]*?<", (("(<ref [^>]*?)<[^<>]*?>", r"\1"),))

        # return the tagged text, without the extra space added at the beginning of this function.
        return text[:-1]

    def extractAllReferences(self, text, tagged=False, splitInChunks=True):
        if splitInChunks:
            allReferences = []
            chunks = text.splitlines(True)
            chunks = [chunks[x:x+self.noOfLinesPerChunkForParsing] for x in range(0, len(chunks), self.noOfLinesPerChunkForParsing)]
            for chunk in chunks:
                allReferences += self.runExtractAllReferences("".join(chunk), tagged)
        else:
            allReferences = self.runExtractAllReferences(text, tagged)
        return allReferences
    
    def runExtractAllReferences(self, text, tagged=False):
        if not tagged:
            text = self.parseText(text, False, True)
        # return a list of tuples (b, c, v) or (b, cs, vs, ce, ve)
        return [literal_eval(m) for m in re.findall('bcv(\([0-9]+?,[ ]*[0-9]+?,[ ]*[0-9, ]*?\))', text)]

    def extractAllReferencesFast(self, text, tagged=False):
        if tagged:
            return "This is not supported"
        sep = ";"
        if "," in text:
            sep = ","
        ret = [self.verseReferenceToBCV(verse) for verse in text.split(sep)]
        return ret

    def parseFile(self, inputFile, splitInChunks=True):
        # set output filename here
        path, file = os.path.split(inputFile)
        outputFile = os.path.join(path, "tagged_{0}".format(file))

        # open file and read input text
        with open(inputFile, "r", encoding="utf-8") as f:
            originalText = f.read()
        newText = self.replaceTextWithReference(originalText, splitInChunks)
        with open(outputFile, "w", encoding="utf-8") as f:
            f.write(newText)

    def replaceTextWithReference(self, originalText, splitInChunks, canonicalOnly=False):
        newText = ""
        if splitInChunks:
            # Create or empty a file
            chunks = originalText.splitlines(True)
            chunks = [chunks[x:x+self.noOfLinesPerChunkForParsing] for x in range(0, len(chunks), self.noOfLinesPerChunkForParsing)]
            for chunk in chunks:
                parsedText = self.runParseText("".join(chunk), canonicalOnly)
                # standardise the format of bible verse references
                if self.standardisation.lower() == "yes":
                    parsedText = self.standardReference(parsedText)
                newText += parsedText
        else:
            parsedText = self.runParseText(originalText, canonicalOnly)
            # standardise the format of bible verse references
            if self.standardisation.lower() == "yes":
                parsedText = self.standardReference(parsedText)
            newText += parsedText
        return newText

    def parseFilesInFolder(self, folder):
        fileList = glob.glob(folder+"/*")
        for file in fileList:
            if os.path.isfile(file):
                self.parseFile(file)

    def extractAllReferencesStartParsing(self, inputName):
        # check if input is a file or a folder
        if os.path.isfile(inputName):
            # parse file
            self.parseFile(inputName)
        elif os.path.isdir(inputName):
            # parse file(s) in a directory
            self.parseFilesInFolder(inputName)
        else:
            # input name is neither a file or a folder
            print("'{0}' is not found.".format(inputName))

    def bookNameToNum(self, bookName):
        bookName = bookName.strip()
        for key in self.sortedNames:
            keyWihtoutDot = re.sub("\.$", "", key)
            if bookName.startswith(keyWihtoutDot):
                return self.bibleBooksDict[key]
        return None

    def verseReferenceToBCV(self, text):
        text = text.strip()
        bible = 0
        for key in self.sortedNames:
            if text.startswith(key):
                bible = self.bibleBooksDict[key]
                break
        reference = text[len(key):]
        res = re.search('(\s*)(\d*):*(\d*) *-* *(\d*):*(\d*)', reference).groups()
        if res[1] == '':
            return (bible, 1, 1)
        elif res[2] == '' and res[3] == '':
            return bible, int(res[1]), 1
        elif res[2] == '' and not res[3] == '':
            return bible, int(res[1]), 1, int(res[3]), 1
        elif res[3] == '' and res[4] == '':
            return bible, int(res[1]), int(res[2])
        elif res[4] == '':
            return bible, int(res[1]), int(res[2]), int(res[1]), int(res[3])
        else:
            return bible, int(res[1]), int(res[2]), int(res[3]), int(res[4])

    # Convert a book string range into a list of book numbers
    # Gen. -> [1]
    # Gen,Exodus,Matt -> [1, 2, 40]
    # Jude-Rev -> [65, 66]
    # Psa, Matt-John, Rev -> [19, 40, 41, 42, 43, 66]
    # OT -> [1, ..., 39]
    # NT -> [40, ..., 66]
    def extractBookList(self, inputString):
        try:
            books = []
            sections = re.split(",", inputString)
            for section in sections:
                if "-" in section:
                    ranges = re.split("-", section)
                    start = self.bookNameToNum(ranges[0])
                    end = self.bookNameToNum(ranges[1])
                    if start is not None and end is not None:
                        for book in range(start, end + 1):
                            books.append(book)
                else:
                    if section.lower() == "ot":
                        for book in range(1, 40):
                            books.append(book)
                    elif section.lower() == "nt":
                        for book in range(40, 67):
                            books.append(book)
                    elif section.lower() in ("all", "clear"):
                        for book in range(1, 500):
                            books.append(book)
                    else:
                        books.append(self.bookNameToNum(section))
            return books
        except:
            return None

    def extractBookListAsString(self, inputString):
        books = self.extractBookList(inputString)
        if books:
            return ",".join([str(book) for book in books if book])
        return ""

    def extractBookListAsBookNameList(self, inputString):
        books = self.extractBookList(inputString)
        bookList = []
        for book in books:
            if book is not None:
                bookList.append(BibleBooks.abbrev["eng"][str(book)][0])
        return bookList

    def extractBookListAsBookNumberList(self, inputString):
        books = self.extractBookList(inputString)
        bookList = []
        for book in books:
            if book is not None:
                bookList.append(int(book))
        return bookList

"""
END - class BibleVerseParser
"""

def extractFromFile():

    inputName = ""
    standardisation = ""
    arguments = sys.argv

    if (len(arguments) > 1):
        inputName = " ".join(arguments[1:])
        standardisation = "NO"
    else:
        # User Interaction - ask for filename / folder name
        inputName = input("Enter a file / folder name here: ")
        # ask if standardising abbreviations and reference format
        standardisation = input("Do you want to standardise the format of all bible verse references? [YES/NO] ")

    # create an instance of BibleVerseParser
    parser = BibleVerseParser(standardisation, noOfLinesPerChunkForParsing=100)
    # start parsing
    parser.extractAllReferencesStartParsing(inputName)

    # delete object
    del parser

def parseImage():
    text = """<html><body>John 1:1, Acts 10:10, <p><img src="data:image/png;base64,iVBOTkSuQmCC" alt="UniqueBibleApp_black" /></p>Rev 22:1</body></html>"""
    result = BibleVerseParser("YES").parseText(text, False, False)
    expectedResult = """<html><body><ref onclick="bcv(43,1,1)">John 1:1</ref>, <ref onclick="bcv(44,10,10)">Acts 10:10</ref>, <p><img src="data:image/png;base64,iVBOTkSuQmCC" alt="UniqueBibleApp_black" /></p><ref onclick="bcv(66,22,1)">Rev 22:1</ref></body></html>"""
    print(result == expectedResult)

if __name__ == '__main__':
    import config
    config.marvelData = "/Users/otseng/dev/UniqueBible/marvelData/"
    from util.ConfigUtil import ConfigUtil
    from util.LanguageUtil import LanguageUtil

    ConfigUtil.setup()
    config.noQt = False
    config.useLiteVerseParsing = True
    config.thisTranslation = LanguageUtil.loadTranslation("en_US")

    bibleVerseParser = BibleVerseParser(config.parserStandarisation)
    print(bibleVerseParser.extractBookList("Jude-Rev"))
    print(bibleVerseParser.extractBookListAsBookNameList("Psa, Matt-John, Rev"))
    print(bibleVerseParser.extractBookListAsBookNumberList("Psa, Matt-John, Rev"))
    print(bibleVerseParser.extractBookListAsString("nt"))
    print(bibleVerseParser.extractBookList("Oadf"))